# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the bin."""
import logging
import os
import unittest
import shutil
import subprocess
import types
import tempfile
import tqdm  # For a progress bar
from io import StringIO
from unittest.mock import patch
from collections.abc import Iterable
from pyretis.bin.pyretisrun import (hello_world,
                                    use_tqdm,
                                    bye_bye_world,
                                    set_up_simulation,
                                    run_generic_simulation,
                                    main,
                                    _RUNNERS)
from pyretis.setup import create_simulation
from pyretis.inout.settings import parse_settings_file

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class test_pyretis_main_runner(unittest.TestCase):
    """Test Main."""

    def test_pyretisrun(self):
        """Test pyretisrun."""
        with patch('sys.stdout', new=StringIO()) as stdout:
            with subprocess.Popen(['pyretisrun', '-V'],
                                  stdout=subprocess.PIPE) as ex:
                asd = ex.stdout.read().split()
            self.assertTrue(b'PyRETIS' in asd)

            with subprocess.Popen(['pyretisrun'],
                                  stderr=subprocess.PIPE) as ex:
                asd = ex.stderr.read().split()
            self.assertTrue(b'pyretisrun' in asd)

    def test_pyretisrun2(self):
        """Test pyretisrun by actually running a retis sim with 0 steps."""
        input_file = 'dummy_input.rst'
        with patch('sys.stdout', new=StringIO()):
            with tempfile.TemporaryDirectory() as tempdir:
                shutil.copyfile(os.path.join(HERE, input_file),
                                os.path.join(tempdir, input_file))
                with subprocess.Popen(['pyretisrun', '-i', input_file, '-p',
                                       '-l', 'DEBUG'],
                                      cwd=tempdir,
                                      stderr=subprocess.DEVNULL,
                                      stdout=subprocess.PIPE) as ex:
                    asd = ex.stdout.read().split()
        self.assertTrue(b'Riccardi' in asd)
        self.assertTrue(b'"kick"' in asd)

    def test_exit(self):
        """Test pyretisrun exit by creating an EXIT file."""
        infile = 'dummy_input.rst'
        with tempfile.TemporaryDirectory() as tempdir:
            input_file = os.path.join(tempdir, infile)
            shutil.copyfile(os.path.join(HERE, infile), input_file)
            with patch('sys.stdout', new=StringIO()) as stdout:
                open(tempdir + '/EXIT', 'w').close()
                main(input_file, tempdir, tempdir,
                     progress=False, log_level=20)
                asd = stdout.getvalue().strip()
        self.assertIn('EXIT file found', asd)

    def test_main(self):
        """Test the main() function for a tis task."""
        infile = 'dummy_input.rst'
        with tempfile.TemporaryDirectory() as tempdir:
            input_file = os.path.join(tempdir, infile)
            shutil.copyfile(os.path.join(HERE, infile), input_file)
            with patch('sys.stdout', new=StringIO()) as stdout:
                main(input_file, tempdir, tempdir, progress=False, log_level=9)
                asd = stdout.getvalue().strip()
            self.assertIn('Execution ended', asd)
            with patch('sys.stdout', new=StringIO()) as stdout:
                input_file = os.path.join(tempdir, 'does_not_exist.rst')
                self.assertRaises(ValueError, main, input_file, tempdir,
                                  tempdir, progress=False,
                                  log_level=9)
                asd = stdout.getvalue().strip()
            self.assertIn('ERROR - execution stopped.', asd)

    def test_run_simulation(self):
        """Test all simulations functions."""
        infile = 'dummy_input.rst'
        runners = _RUNNERS
        runners['umbrellawindow'] = run_generic_simulation
        for key, runner in runners.items():
            with patch('sys.stdout', new=StringIO()):
                with tempfile.TemporaryDirectory() as tempdir:
                    if key in ['repptis', 'retis']:
                        continue
                    input_file = os.path.join(tempdir, infile)
                    shutil.copyfile(os.path.join(HERE, infile), input_file)
                    sim_settings = parse_settings_file(input_file)
                    sim_settings['simulation']['exe_path'] = tempdir
                    sim_settings['engine']['exe_path'] = tempdir
                    for ens in sim_settings['ensemble']:
                        ens['simulation']['exe_path'] = tempdir
                        ens['engine']['exe_path'] = tempdir
                    sim_settings['simulation']['task'] = key
                    sim = create_simulation(sim_settings)
                    self.assertTrue(runner(sim, sim_settings))

    def test_simulation_exit(self):
        """Test exit feature of explore_simulation and run_path_simulation."""
        infile = 'dummy_input.rst'
        runners = _RUNNERS
        for key, runner in runners.items():
            with patch('sys.stdout', new=StringIO()):
                with tempfile.TemporaryDirectory() as tempdir:
                    if key not in ('tis', 'explore'):
                        continue
                    input_file = os.path.join(tempdir, infile)
                    shutil.copyfile(os.path.join(HERE, infile), input_file)
                    sim_settings = parse_settings_file(input_file)
                    sim_settings['simulation']['exe_path'] = tempdir
                    sim_settings['engine']['exe_path'] = tempdir
                    for ens in sim_settings['ensemble']:
                        ens['simulation']['exe_path'] = tempdir
                        ens['engine']['exe_path'] = tempdir
                    sim_settings['simulation']['task'] = key
                    sim = create_simulation(sim_settings)
                    open(tempdir + '/EXIT', 'w').close()
                    self.assertFalse(runner(sim, sim_settings))

    def test_use_tqdm(self):
        """Test tqdm."""
        bar = use_tqdm(progress=True)
        self.assertEqual(type(bar), type(tqdm.tqdm))
        self.assertTrue(bar is tqdm.tqdm)
        self.assertFalse(isinstance(bar, Iterable))

        bar = use_tqdm(progress=False)
        self.assertTrue(bar is not tqdm.tqdm)
        self.assertTrue(isinstance(bar, types.FunctionType))
        value = bar(iterable='something1', dummy='something2')
        self.assertEqual(value, 'something1')

    def test_hello_world(self):
        """Test that we are polite."""
        with patch('sys.stdout', new=StringIO()) as stdout:
            hello_world(infile='I_can_read_your_mind.rst',
                        rundir=HERE,
                        logfile='nothing_to_declare')
        self.assertIn('Start', stdout.getvalue().strip())

    def test_bye_world(self):
        """Test that we can die."""
        with patch('sys.stdout', new=StringIO()) as stdout:
            bye_bye_world()
        self.assertIn('reference', stdout.getvalue().strip())

    def test_set_up_simulation(self):
        """Test that we know how to set up a simulation."""
        inputfile = 'a_non_existent_input.rst'
        with self.assertRaises(ValueError) as err:
            set_up_simulation(inputfile, HERE)
        self.assertEqual(f'Input file "{inputfile}" NOT found!',
                         str(err.exception))

        inputfile = os.path.join(HERE, 'dummy_input.rst')
        with patch('sys.stdout', new=StringIO()) as stdout:
            set_up_simulation(inputfile, HERE)
        self.assertIn('Reading', stdout.getvalue().strip())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Functions for generating plots using matplotlib.

This module defines a plotter class for matplotlib and it also defines
some standard plots that are used in the analysis.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MplPlotter (:py:class:`.MplPlotter`)
    A class for plotting with matplotlib.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

mpl_set_style (:py:func:`.mpl_set_style`)
    Method for setting the style for the plots, typically used here to
    load the *PyRETIS style*.
"""
import os
import logging
import numpy as np
import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.collections import LineCollection
import matplotlib.style
from pyretis.inout.plotting.plotting import Plotter
from pyretis.inout.common import create_backup, name_file
from pyretis.inout.common import (ENERFILES, ENERTITLE, FLUXFILES,
                                  ORDERFILES, PATHFILES, PATH_MATCH)

np.set_printoptions(legacy='1.21')

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['MplPlotter']


# Define default style file:
_STYLEFILE = 'pyretis.mplstyle'
_MPL_STYLE_FILE = os.sep.join([os.path.dirname(__file__), 'styles',
                               _STYLEFILE])
_TITLE_SETTINGS = {'loc': 'right'}


class MplPlotter(Plotter):
    """A plotter using matplotlib.

    This class defines a plotter which makes use of matplotlib and
    it can be used as a starting point to create other plotters
    for PyRETIS bases on other tools (e.g. gnuplot etc.).

    Attributes
    ----------
    style : string
        Defines what style to use for the plotting.
    out_fmt : string
        Selects format for output plots.

    """

    def __init__(self, out_fmt, backup=False, style=None, out_dir=None):
        """Initiate, set style and check output format.

        Parameters
        ----------
        out_fmt : string
            This is the format to use for the images.
        backup : boolean, optional
            Determines if we should overwrite or backup old files.
        style : string, optional
            This selects the style to use, it can be a file path or the
            string with the style name.
        out_dir : string, optional
            Determines if we should write the files to a particular
            directory.

        """
        super().__init__(backup=backup, plotter_type='matplotlib',
                         out_dir=out_dir)
        self.style = style
        mpl_set_style(self.style)
        # Check if the requested file format is something we can do:
        fig = Figure()
        canvas = FigureCanvas(fig)
        supported = canvas.get_supported_filetypes()
        if out_fmt not in supported:
            msg = [f'Output format "{out_fmt}" is not supported.',
                   'Please try:']
            for key in supported:
                msg.append(key)
            raise ValueError(' '.join(msg))
        self.out_fmt = out_fmt
        del fig
        del canvas

    def _print_figures_to_file(self, canvas):
        """Save figure(s) to file(s).

        This function will save figures to files. It will add a file
        format extension to the files when writing.

        Parameters
        ----------
        canvas : dict of :class:`matplotlib.backend_bases.FigureCanvasBase`
            ``canvas[key]`` is assumed to define a figure which we will
            save to a file with a file name defined by ``key`` and the
            extension defined by ``self.out_fmt``.

        Returns
        -------
        out : dict
            The files written.

        """
        outputfiles = {}
        for key, val in canvas.items():
            local_file = name_file(key, self.out_fmt, path=None)
            full_path = name_file(key, self.out_fmt, self.out_dir)
            outputfiles[key] = local_file
            mpl_savefig(val, full_path, self.backup)
        return outputfiles

    def output_flux(self, results):
        """Plot flux results using py:func:`.mpl_plot_flux`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_flux`.

        Returns
        -------
        outputfile : list of dicts
            A list containing the files created for the flux and for
            the error in the flux.

        Note
        ----
        The returned list is used to plot the figures in *pairs*.

        """
        canvas_run, canvas_err = mpl_plot_flux(results)
        # Restructure output files for reporting
        outputfiles = []
        for run, err in zip(canvas_run, canvas_err):
            local_run = name_file(run['name'], self.out_fmt, path=None)
            local_err = name_file(err['name'], self.out_fmt, path=None)
            full_run = name_file(run['name'], self.out_fmt, path=self.out_dir)
            full_err = name_file(err['name'], self.out_fmt, path=self.out_dir)
            mpl_savefig(run['canvas'], full_run, self.backup)
            mpl_savefig(err['canvas'], full_err, self.backup)
            outputfiles.append({'runflux': local_run,
                                'errflux': local_err})
        return outputfiles

    def output_energy(self, results, energies):
        """Plot energy results using :py:func:`.mpl_plot_energy`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_energy`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas = mpl_plot_energy(results, energies)
        return self._print_figures_to_file(canvas)

    def output_orderp(self, results, orderdata):
        """Plot order parameter data using :py:func:`.mpl_plot_orderp`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_orderp`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas = mpl_plot_orderp(results, orderdata)
        return self._print_figures_to_file(canvas)

    def output_path(self, results, path_ensemble):
        """Plot path results using :py:func:`.mpl_plot_path`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_path`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas = mpl_plot_path(results, path_ensemble)
        return self._print_figures_to_file(canvas)

    def output_pppath(self, results, path_ensemble):
        """Plot repptis path results using :py:func:`.mpl_plot_pppath`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_pppath`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas = mpl_plot_pppath(results, path_ensemble)
        return self._print_figures_to_file(canvas)

    def output_pp_global_cross(self, pp):
        """Plot repptis global cross prob using :py:func:`.mpl_plot_path`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_path`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        series = [{'type': 'xy',
                   'x': list(range(len(pp))),
                   'y': pp}]
        figset = {'xlabel': r'$\lambda_i$',
                  'ylabel': r'$P_A^{(\lambda_i|\lambda_A)}$',
                  'title': 'Global crossing probability', 'yscale': 'log'}
        canvas = mpl_simple_plot(series, fig_settings=figset)
        return self._print_figures_to_file({'pp_Pc': canvas})

    def output_matched_probability(self, path_ensembles, detect,
                                   matched, reptis=False):
        """Plot matched probabilities using :py:func:`.mpl_plot_matched`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_matched`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas = mpl_plot_matched(path_ensembles, detect, matched, reptis)
        return self._print_figures_to_file(canvas)

    def output_xi(self, xi, xierror):
        """Plot Xi and Xi error using :py:func:`.mpl_plot_xi`.

        The parameters for this method is described in
        :py:func:`.mpl_plot_xi`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas_run, canvas_err = mpl_plot_xi(xi, xierror)
        local_run = name_file(canvas_run[0]['name'], self.out_fmt, path=None)
        full_run_path = name_file(canvas_run[0]['name'],
                                  self.out_fmt,
                                  self.out_dir)
        mpl_savefig(canvas_run[0]['canvas'], full_run_path, self.backup)

        local_err = name_file(canvas_err[0]['name'], self.out_fmt, path=None)
        full_err_path = name_file(canvas_err[0]['name'],
                                  self.out_fmt,
                                  self.out_dir)
        mpl_savefig(canvas_err[0]['canvas'], full_err_path, self.backup)

        return {"xirun": local_run, "xierror": local_err}

    def output_tau(self, tau, tauerror, bins, tau_ref, tau_ref_bins):
        """Plot Tau, Tau error, and tau reference histogram.

        The parameters for this method is described in
        :py:func:`.mpl_plot_tau` and :py:func:`.mpl_plot_tau_ref`.

        Returns
        -------
        out : dict
            This dict contains the files created by the plotting.

        """
        canvas_run, canvas_err = mpl_plot_tau(tau, tauerror)
        canvas_ref = mpl_plot_tau_ref(tau_ref, tau_ref_bins, bins)
        local_run = name_file(canvas_run[0]['name'], self.out_fmt, path=None)
        full_run_path = name_file(canvas_run[0]['name'],
                                  self.out_fmt,
                                  self.out_dir)
        mpl_savefig(canvas_run[0]['canvas'], full_run_path, self.backup)

        local_err = name_file(canvas_err[0]['name'], self.out_fmt, path=None)
        full_err_path = name_file(canvas_err[0]['name'],
                                  self.out_fmt,
                                  self.out_dir)
        mpl_savefig(canvas_err[0]['canvas'], full_err_path, self.backup)
        local_ref = name_file(canvas_ref['name'], self.out_fmt, path=None)
        full_ref_path = name_file(canvas_ref['name'],
                                  self.out_fmt,
                                  self.out_dir)
        mpl_savefig(canvas_ref['canvas'], full_ref_path, self.backup)
        return {"taurun": local_run, "tauerror": local_err,
                "tauref": local_ref}


def _mpl_read_style_file(filename):
    """Read style files for old versions of matplotlib.

    This function is just intended for old versions of matplotlib
    where we have to read parameters and set them ourselves.

    Parameters
    ----------
    filename : string
        This is the matplotlib rc file to open.

    Returns
    -------
    out : None
        Returns None, but modifies `matplotlib.rcParams`.

    """
    with open(filename, 'r', encoding="utf8") as fileh:
        for lines in fileh:
            linesc = lines.strip().split('#')[0]
            loc = linesc.find(':')
            if loc != -1:
                key = linesc[:loc].strip()
                value = linesc[loc+1:].strip()
                if key.find('color') != -1:
                    value = f'#{value}'
                try:
                    matplotlib.rcParams[key] = value
                except KeyError:
                    logger.warning(('Unknown setting "%s". '
                                    'Please update matplotlib'), key)


def mpl_set_style(style='pyretis'):
    """Set the plotting style for matplotlib.

    This will set up the plotting according to some given style.
    Styles can be given as string, for instance 'ggplot', 'bmh',
    'grayscale' (i.e. one of the styles in `matplotlib.style.available`)
    or as a file (full path is needed). The default PyRETIS style
    is stored in `_MPL_STYLE_FILE` and can be selected with 'pyretis'.
    Style equal to None is just the default matplotlib style.

    Parameters
    ----------
    style : string, optional
        This selects the style to use, it can be a file path or the
        string with the style name.

    """
    if style == 'pyretis':
        style = _MPL_STYLE_FILE
    if style in matplotlib.style.available:
        logger.info('Loading matplotlib style: %s', style)
        matplotlib.style.use(style)
    else:  # assume this is just a file
        logger.info('Loading matplotlib style from file: %s', style)
        rcpar = matplotlib.rc_params_from_file(
            style,
            use_default_template=False
        )
        matplotlib.rcParams.update(rcpar)


def mpl_savefig(canvas, outputfile, backup=False):
    """Write/save matplotlib figures to files.

    Parameters
    ----------
    canvas : object like :class:`matplotlib.backend_bases.FigureCanvasBase`
        This is the figure to be written to the file by
        using ``canvas.print_figure()``.
    outputfile : string
        This is the name of the output file to create.
    backup : boolean, optional
        This determines if we should try to back-up old versions of the
        figures.

    """
    if backup:
        msg = create_backup(outputfile)
        if msg:
            logger.warning(msg)
    canvas.print_figure(outputfile)


def mpl_plot_in_chunks(axs, series, chunksize=20000):
    """Plot a series in chunks using matplotlib.

    When plotting 'large' datasets, matplotlib might give an
    ``OverflowError: Allocated too many blocks`` error.
    Here we avoid this error by plotting the data in chunks. We could
    also downsample the data, but this is perhaps something best left
    to the user.

    Parameters
    ----------
    axs : Axes object from matplotlib
        Where to do the plotting
    series : dict
        Represents the data to be plotted.
    chunksize : int
        This is the maximum size we will try to plot in one go.

    """
    color = None
    line = None
    leny = min(len(series['y']), len(series['x']))
    if leny > chunksize:
        nchunk, rest = divmod(leny, chunksize)
        for i in range(nchunk):
            low = i * chunksize
            high = low + chunksize
            line = _mpl_plot_xy_chunk(axs, series, low=low, high=high,
                                      color=color)
            color = line.get_color()
        if rest > 0:
            # We want 1 less to force line overlap,
            # We want positive low in case x and y length don't overlap
            line = _mpl_plot_xy_chunk(axs, series, low=nchunk*chunksize-1,
                                      high=None, color=color)
    else:
        line = _mpl_plot_xy_chunk(axs, series)
    return line


def _mpl_plot_xy_chunk(axs, series, low=0, high=None, color=None):
    """Do the actual plotting in chunks for ``x`` vs ``y`` data.

    Parameters
    ----------
    axs : Axes object from matplotlib
        Where to do the plotting.
    series : dict
        Represents the data to be plotted.
    low : int, optional
        Lower index to start plotting. `low` can be negative.
    high : int, optional
        Index where to end the plotting, this index is not plotted.
        `high` is assumed to always be > 0 or None.
    color : string, optional
        A string representing the color to use.

    Returns
    -------
    handle : object like :py:class:`matplotlib.lines.Line2D`
        A handle for the plotted line.

    """
    # pick out just a few keys - we want to limit what we change here:
    kwargs = {'linestyle': series.get('ls', '-'),
              'alpha': series.get('alpha', 1.0),
              'linewidth': series.get('lw', 2.0)}
    if color is not None:
        kwargs['color'] = color
    elif 'color' in series:
        kwargs['color'] = series['color']

    if high is None:
        high = min(len(series['x']), len(series['y']))
    else:
        high = min(len(series['x']), len(series['y']), high)
    handle, = axs.plot(series['x'][low:high], series['y'][low:high],
                       **kwargs)
    return handle


def mpl_simple_plot(series, fig_settings=None):
    """Plot simple time series data (i.e. ``x`` vs ``y`` data).

    Parameters
    ----------
    series : list of dicts
        `series[i]` is the dict which contains the data to be plotted.
    fig_settings : dict, optional
        This dict contains settings for the figure, keys are:

        * `xlabel`: string, the label to use for the x-axis.
        * `ylabel`: string, the label to use for the y-axis.
        * `title`: string, title to use for the figure.
        * `yscale`: string, to change the scale for the y-axis.

    Returns
    -------
    out : object like :class:`matplotlib.backend_bases.FigureCanvasBase`
        This is the figure we create here.

    """
    fig = Figure()
    canvas = FigureCanvas(fig)
    axs = fig.add_subplot(111)
    handles = []
    labels = []
    for seri in series:
        handle = None
        if seri['type'] == 'xy':
            handle = mpl_plot_in_chunks(axs, seri)
        elif seri['type'] == 'bar':
            handle = mpl_plot_bar(axs, seri)
        elif seri['type'] == 'vline':
            handle = axs.axvline(x=seri['x'], ls=seri.get('ls', '-'),
                                 alpha=seri.get('alpha', 1.0),
                                 lw=seri.get('lw', 2.0))
        elif seri['type'] == 'hline':
            handle = axs.axhline(y=seri['y'], ls=seri.get('ls', '-'),
                                 alpha=seri.get('alpha', 1.0),
                                 lw=seri.get('lw', 2.0))
        legend = seri.get('label', None)
        if legend is not None and handle is not None:
            handles.append(handle)
            labels.append(legend)
    if 'xlabel' in fig_settings:
        axs.set_xlabel(fig_settings['xlabel'])
    if 'ylabel' in fig_settings:
        axs.set_ylabel(fig_settings['ylabel'])
    if 'title' in fig_settings:
        axs.set_title(fig_settings['title'], **_TITLE_SETTINGS)
    if len(labels) == len(handles) and len(labels) >= 1:
        ncol, rest = divmod(len(labels), 10)
        if rest > 0:
            ncol += 1
        if ncol <= 2:
            axs.legend(handles, labels, ncol=ncol)
    if 'yscale' in fig_settings:
        axs.set_yscale(fig_settings['yscale'])
    return canvas


def mpl_linecollection_gradient(axs, series):
    """Plot a line with a color gradient along the line.

    The line is split into segments and each segment is colored
    according to it's "location" along the line. By "location" we here
    mean the order of the segments.

    Parameters
    ----------
    axs : Axes object from matplotlib
        Where to do the plotting.
    series : dict
        Represents the data to be plotted.

    Returns
    -------
    handle : object like :class:`matplotlib.collections.LineCollection`
        A handle for the plotted line.

    """
    # pick out just a few keys - we want to limit what we change here:
    kwargs = {'linestyle': series.get('ls', '-'),
              'alpha': series.get('alpha', 1.0),
              'linewidth': series.get('lw', 2.0)}
    points = np.array([series['x'], series['y']]).T.reshape((-1, 1, 2))
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    values = np.linspace(0, 1, len(series['x']))
    linec = LineCollection(segments, array=values,
                           norm=matplotlib.colors.Normalize(vmin=0, vmax=1),
                           **kwargs)
    handle = axs.add_collection(linec)
    return handle


def mpl_chunks_gradient(axs, series, chunksize=20000):
    """Plot a line gradient in chunks.

    Here we will plot a line in chunks and color each chunk with one
    color. This function can be used as an alternative to
    :py:func:`.mpl_linecollection_gradient` when the number of points
    to plot is very large. Typically the chunk size here will be small
    compared to the size of the data to be plotted, so that each chunk,
    if plotted with :py:func:`.mpl_linecollection_gradient`, would have
    approximately the same color anyway.

    Parameters
    ---------
    axs : Axes object from matplotlib
        Where to do the plotting.
    series : dict
        Represents the data to be plotted.
    chunksize : int, optional
        This is the maximum size (number of points) we will try to plot
        in one go.

    Returns
    -------
    handle : object like :class:`matplotlib.lines.Line2D`
        A handle for the plotted line.

    Note
    ----
    Color maps in matplotlib will typically have 256 colors. The number
    of different colors we can get is currently limited to 256.

    """
    kwargs = {'linestyle': series.get('ls', '-'),
              'alpha': series.get('alpha', 1.0),
              'linewidth': series.get('lw', 2.0)}
    line = None
    lenx = len(series['x'])
    nchunk, rest = divmod(lenx, chunksize)
    cnorm = matplotlib.colors.Normalize(vmin=0, vmax=nchunk+1)
    color_map = matplotlib.cm.ScalarMappable(norm=cnorm)
    for i in range(nchunk):
        low = i * chunksize
        high = low + chunksize
        line, = axs.plot(series['x'][low:high], series['y'][low:high],
                         **kwargs)
        line.set_color(color_map.to_rgba(i))
    if rest > 0:
        last = rest + 1
        line, = axs.plot(series['x'][-last:], series['y'][-last:],
                         **kwargs)
        line.set_color(color_map.to_rgba(nchunk))
    return line


def mpl_line_gradient(series, fig_settings):
    """Plot time series with a color gradient.

    This function will plot time series data and color the lines with
    a gradient according to 'time'

    Parameters
    ----------
    series : list of dicts
        `series[i]` is the dict which contains the data to be plotted.
    fig_settings : dict
        This dict contains settings for the figure, keys are:

        * `xlabel`: string, the label to use for the x-axis.
        * `ylabel`: string, the label to use for the y-axis.
        * `title`: string, title to use for the figure.
        * `yscale`: string, to change the scale for the y-axis.

    Returns
    -------
    out : object like :class:`matplotlib.backend_bases.FigureCanvasBase`
        This is the figure we create here.

    Notes
    -----
    This function is based on the matplotlib example from:
    http://matplotlib.org/examples/pylab_examples/multicolored_line.html

    """
    fig = Figure()
    canvas = FigureCanvas(fig)
    axs = fig.add_subplot(111)
    handles = []
    labels = []
    for seri in series:
        lenx = len(seri['x'])
        if lenx >= 10**6:  # plot in chunks
            logger.info('Line gradient: LARGE dataset - plotting in chunks')
            handle = mpl_chunks_gradient(axs, seri)
        else:  # just plot it all
            handle = mpl_linecollection_gradient(axs, seri)
        legend = seri.get('label', None)
        if legend is not None and handle is not None:
            handles.append(handle)
            labels.append(legend)
    axs.autoscale_view()
    if 'xlabel' in fig_settings:
        axs.set_xlabel(fig_settings['xlabel'])
    if 'ylabel' in fig_settings:
        axs.set_ylabel(fig_settings['ylabel'])
    if 'title' in fig_settings:
        axs.set_title(fig_settings['title'], **_TITLE_SETTINGS)
    if len(labels) == len(handles) and len(labels) >= 1:
        axs.legend(handles, labels)
    return canvas


def mpl_error_plot(series, fig_settings):
    """Plot series with errors.

    Plot a series with error values displayed as a filled region.

    Parameters
    ----------
    series : list of tuples
        `series[i]` is the tuple which will be plotted. It is assumed
        to be on the form (x-values, y-values, y-error, legend).
    fig_settings : dict
        This dict contains settings for the figure, keys are:

        * `xlabel`: string, the label to use for the x-axis.
        * `ylabel`: string, the label to use for the y-axis.
        * `title`: string, title to use for the figure.

    Returns
    -------
    out : object like :class:`matplotlib.backend_bases.FigureCanvasBase`
        This is the figure we create here.

    """
    fig = Figure()
    canvas = FigureCanvas(fig)
    axs = fig.add_subplot(111)
    handles = []
    labels = []
    for seri in series:
        try:
            add_legend = seri[3] is not None
        except IndexError:
            add_legend = False
        handle, = axs.plot(seri[0], seri[1])
        axs.fill_between(seri[0], seri[1] + seri[2],
                         seri[1] - seri[2],
                         facecolor=handle.get_color(), alpha=0.3)
        if add_legend:
            handles.append(handle)
            labels.append(seri[3])
    if 'xlabel' in fig_settings:
        axs.set_xlabel(fig_settings['xlabel'])
    if 'ylabel' in fig_settings:
        axs.set_ylabel(fig_settings['ylabel'])
    if 'title' in fig_settings:
        axs.set_title(fig_settings['title'], **_TITLE_SETTINGS)
    if len(labels) == len(handles) and len(labels) >= 1:
        axs.legend(handles, labels)
    return canvas


def _mpl_shoots_histogram(histograms, ensemble):
    """Plot the histograms from the shoots analysis.

    Parameters
    ----------
    histograms : dict
        These are the histograms obtained in the shoots analysis.
    ensemble : string
        This is the ensemble identifier, e.g. 001, 002, etc.

    Returns
    -------
    out : object like :class:`matplotlib.backend_bases.FigureCanvasBase`
        This is the unscaled histogram.

    """
    series = []
    for key in ['ACC']:
        try:
            mid = histograms[key][2]
            hist = histograms[key][0]
            series.append({'type': 'xy', 'x': mid, 'y': hist,
                           'alpha': 0.8})
        except KeyError:
            continue
    figset = {'xlabel': 'Order parameter', 'ylabel': 'Probability density',
              'title': f'Ensemble ${ensemble}$'}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    return canvas


def mpl_plot_path(results, path_ensemble):
    """Plot all figures from the path analysis.

    Parameters
    ----------
    results : dict
        This dict contains the result from the analysis.
    path_ensemble : object like :py:class:`.PathEnsemble`
        This is the path ensemble we have analyzed.

    Returns
    -------
    canvas : dict
        This dictionary contains the different canvases we have
        created.

    """
    ens = path_ensemble.ensemble_name
    ens_simplified = path_ensemble.ensemble_name_simple
    detect = results['detect']
    canvas = {}
    out = {}
    for key, item in PATHFILES.items():
        out[key] = item.format(ens_simplified)
    if 'pcross' in results:
        # First plot `pcross` vs `lambda` with the `detect` surface:
        series = [
            {'type': 'xy',
             'x': results['pcross'][0],
             'y': results['pcross'][1]},
            {'type': 'vline',
             'x': detect,
             'ls': '--',
             'alpha': 0.8},
        ]
        figset = {'xlabel': r'Order parameter ($\lambda$)',
                  'ylabel': 'Crossing probability',
                  'title': f'Ensemble ${ens}$'}
        canvas[out['pcross']] = mpl_simple_plot(series, fig_settings=figset)
    if 'prun' in results:
        # Next plot running ` pcross`:
        series = [
            {'type': 'xy',
             'x': results['cycle'],
             'y': results['prun']},
            {'type': 'hline',
             'y': results['prun'][-1],
             'ls': '--',
             'alpha': 0.8},
        ]
        figset = {'xlabel': 'Cycle number',
                  'ylabel': 'Running estimate',
                  'title': f'Ensemble ${ens}$'}
        canvas[out['prun']] = mpl_simple_plot(series, fig_settings=figset)
    if 'blockerror' in results:
        # Plot results of block-error analysis:
        series = [
            {'type': 'xy',
             'x': results['blockerror'][0],
             'y': results['blockerror'][3]},
            {'type': 'hline',
             'y': results['blockerror'][4],
             'ls': '--',
             'alpha': 0.8},
        ]
        title = r'Ensemble ${0}$: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
                  'title': title.format(ens, results['blockerror'][4],
                                        results['blockerror'][6])}
        canvas[out['perror']] = mpl_simple_plot(series, fig_settings=figset)
    if 'lengtherror' in results:
        # Plot results of length-error analysis:
        series = [
            {'type': 'xy',
             'x': results['lengtherror'][0],
             'y': results['lengtherror'][3]},
            {'type': 'hline',
             'y': results['lengtherror'][4],
             'ls': '--',
             'alpha': 0.8},
        ]
        title = r'Ensemble ${0}$: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
                  'title': title.format(ens, results['lengtherror'][4],
                                        results['lengtherror'][6])}
        canvas[out['lengtherror']] = mpl_simple_plot(series,
                                                     fig_settings=figset)

    # Plot length-histogram:
    labfmt = r'Avg. {0:6.2f} st.dev. {1:6.2f}'
    series = [
        {'type': 'xy',
         'x': results['pathlength'][0][1],
         'y': results['pathlength'][0][0],
         'label': labfmt.format(results['pathlength'][0][2][0],
                                results['pathlength'][0][2][1])},
    ]
    figset = {'xlabel': 'No. of PyRETIS engine steps', 'ylabel': 'Probability',
              'title': f'Ensemble ${ens}$'}
    canvas[out['pathlength']] = mpl_simple_plot(series, fig_settings=figset)
    # Plot shoots-histogram
    can_tmp = _mpl_shoots_histogram(results['shoots'][0], ens)
    canvas[out['shoots']] = can_tmp
    return canvas


def mpl_plot_pppath(results, path_ensemble):
    """Plot all figures from the repptis path analysis.

    Parameters
    ----------
    results : dict
        This dict contains the result from the analysis.
    path_ensemble : object like :py:class:`.PathEnsemble`
        This is the path ensemble we have analyzed.

    Returns
    -------
    canvas : dict
        This dictionary contains the different canvases we have
        created.

    """
    ens = path_ensemble.ensemble_name
    ens_simplified = path_ensemble.ensemble_name_simple
    canvas = {}
    out = {}
    colors = get_color_map(10)
    for key, item in PATHFILES.items():
        out[key] = item.format(ens_simplified)
    if 'pp_Pc' in results:
        # First plot `pcross` vs `lambda` with the `detect` surface:
        series = [
            {'type': 'xy',
             'x': list(range(len(results['pp_Pc']))),
             'y': results['pp_Pc']},
        ]
        figset = {'xlabel': r'intf',
                  'ylabel': r'$P_a(\lambda_i/\lambda_A)$',
                  'title': r''}
        canvas[out['pp_Pc']] = mpl_simple_plot(series, fig_settings=figset)
    if 'pcross' in results:
        L, M, R, lmlpercs, lmllambs, rmrpercs, rmrlambs = results['repptis']
        # First plot `pcross` vs `lambda` with the `detect` surface:
        series = [
            {'type': 'xy',
             'x': lmllambs,
             'y': lmlpercs,
             'color': colors[0]},
            {'type': 'xy',
             'x': rmrlambs,
             'y': rmrpercs,
             'color': colors[1]},
            {'type': 'vline',
             'x': L,
             'ls': '--',
             'alpha': 0.8},
            {'type': 'vline',
             'x': M,
             'ls': '--',
             'alpha': 0.8},
            {'type': 'vline',
             'x': R,
             'ls': '--',
             'alpha': 0.8},
        ]
        figset = {'xlabel': r'Order parameter ($\lambda$)',
                  'ylabel': 'Crossing probability',
                  'title': f'Ensemble ${ens}$'
                           + f', L = {L}, M = {M}, R = {R}'}
        canvas[out['pcross']] = mpl_simple_plot(series, fig_settings=figset)
    if 'prun_sl' in results:
        # Next plot running ` pcross`:
        series = [
            {'type': 'xy',
             'x': results['cycle'],
             'y': results['prun_sl'],
             'color': colors[0]},
            {'type': 'hline',
             'y': results['prun_sl'][-1],
             'ls': '--',
             'alpha': 0.8},
        ]
        figset = {'xlabel': 'Cycle number',
                  'ylabel': 'Running estimate',
                  'title': f'Ensemble ${ens}$'
                           + f", {results['prun_sl'][-1]}"}
        canvas[out['prun_sl']] = mpl_simple_plot(series, fig_settings=figset)
    if 'prun_sr' in results:
        # Next plot running ` pcross`:
        series = [
            {'type': 'xy',
             'x': results['cycle'],
             'y': results['prun_sr'],
             'color': colors[1]},
            {'type': 'hline',
             'y': results['prun_sr'][-1],
             'ls': '--',
             'alpha': 0.8},
        ]
        figset = {'xlabel': 'Cycle number',
                  'ylabel': 'Running estimate',
                  'title': f'Ensemble ${ens}$'
                           + f", {results['prun_sr'][-1]}"}
        canvas[out['prun_sr']] = mpl_simple_plot(series, fig_settings=figset)
    if 'blockerror_sl' in results:
        # Plot results of block-error analysis:
        series = [
            {'type': 'xy',
             'x': results['blockerror_sl'][0],
             'y': results['blockerror_sl'][3],
             'color': colors[0]},
            {'type': 'hline',
             'y': results['blockerror_sl'][4],
             'ls': '--',
             'alpha': 0.8},
        ]
        title = r'Ensemble ${0}$: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
                  'title': title.format(ens, results['blockerror_sl'][4],
                                        results['blockerror_sl'][6])}
        canvas[out['perror_sl']] = mpl_simple_plot(series, fig_settings=figset)
    if 'blockerror_sr' in results:
        # Plot results of block-error analysis:
        series = [
            {'type': 'xy',
             'x': results['blockerror_sr'][0],
             'y': results['blockerror_sr'][3],
             'color': colors[1]},
            {'type': 'hline',
             'y': results['blockerror_sr'][4],
             'ls': '--',
             'alpha': 0.8},
        ]
        title = r'Ensemble ${0}$: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
                  'title': title.format(ens, results['blockerror_sr'][4],
                                        results['blockerror_sr'][6])}
        canvas[out['perror_sr']] = mpl_simple_plot(series, fig_settings=figset)
    if 'lengtherror' in results:
        # Plot results of length-error analysis:
        series = [
            {'type': 'xy',
             'x': results['lengtherror'][0],
             'y': results['lengtherror'][3]},
            {'type': 'hline',
             'y': results['lengtherror'][4],
             'ls': '--',
             'alpha': 0.8},
        ]
        title = r'Ensemble ${0}$: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
                  'title': title.format(ens, results['lengtherror'][4],
                                        results['lengtherror'][6])}
        canvas[out['lengtherror']] = mpl_simple_plot(series,
                                                     fig_settings=figset)

    # Plot length-histogram:
    labfmt = r'Avg8 {0:6.2f} st.dev. {1:6.2f}'
    series = [
        {'type': 'xy',
         'x': results['pathlength'][0][1],
         'y': results['pathlength'][0][0],
         'label': labfmt.format(results['pathlength'][0][2][0],
                                results['pathlength'][0][2][1])},
    ]
    figset = {'xlabel': 'No. of PyRETIS engine steps',
              'ylabel': 'Crossing probability', 'title': f'Ensemble ${ens}$'}
    canvas[out['pathlength']] = mpl_simple_plot(series, fig_settings=figset)
    # Plot shoots-histogram
    can_tmp = _mpl_shoots_histogram(results['shoots'][0], ens)
    canvas[out['shoots']] = can_tmp
    return canvas


def mpl_plot_orderp(results, orderdata):
    """Plot the output from the order parameter analysis using matplotlib.

    Parameters
    ----------
    results : dict
        Each item in `results` contains the results for the
        corresponding order parameter.
    orderdata : list of numpy.arrays
        This is the raw data for the order parameter analysis.

    Returns
    -------
    canvas : dict
        The different plots created by this function.

    Note
    ----
    We are here only outputting results for the first order parameter.
    I.e. other order parameters or velocities are not written here. This
    will be changed when the structure of the output order parameter
    file has been fixed. Also note that, if present, the first order
    parameter will be plotted against the second one - i.e. the second
    one will be assumed to represent the velocity here.

    """
    canvas = {}
    time = orderdata[:, 0]
    series = [{'type': 'xy', 'x': time, 'y': orderdata[:, 1]}]
    figset = {'xlabel': 'Time', 'ylabel': 'Order parameter'}
    canvas[ORDERFILES['order']] = mpl_simple_plot(series, fig_settings=figset)
    # make running average plot of the energies as function of time
    series = [{'type': 'xy', 'x': time, 'y': results[0]['running'],
               'label': 'Running average'}]
    canvas[ORDERFILES['run_order']] = mpl_simple_plot(series,
                                                      fig_settings=figset)
    # plot block-error results:
    block = results[0]['blockerror']
    series = [
        {'type': 'xy', 'x': block[0], 'y': block[3]},
        {'type': 'hline', 'y': block[4], 'ls': '--', 'alpha': 0.8}
    ]
    title = 'Order parameter. Rel. err.: {0:9.6e}, Ncor: {1:9.6f}'
    figset = {'xlabel': 'Block length',
              'ylabel': 'Estimated error',
              'title': title.format(block[4], block[6])}
    canvas[ORDERFILES['block']] = mpl_simple_plot(series, fig_settings=figset)
    # plot distributions
    dist = results[0]['distribution']
    series = [{'type': 'xy', 'x': dist[1], 'y': dist[0]}]
    title = '{0}. Average: {1:9.6e}, std: {2:9.6f}'
    title = title.format('Order parameter', dist[2][0], dist[2][1])
    figset = {'title': title}
    canvas[ORDERFILES['dist']] = mpl_simple_plot(series,
                                                 fig_settings=figset)
    # also try a orderp vs ordervel plot:
    _, col = orderdata.shape
    if col >= 3:
        series = [{'type': 'xyc', 'x': orderdata[:, 1], 'y': orderdata[:, 2]}]
        figset = {'xlabel': r'$\lambda$',
                  'ylabel': r'$\dot{\lambda}$',
                  'title': 'Order parameter vs velocity'}
        plot = ORDERFILES['ordervel']
        canvas[plot] = mpl_line_gradient(series, fig_settings=figset)
    # output msd if it was calculated:
    if 'msd' in results[0]:
        msd = results[0]['msd']
        series = [(np.arange(len(msd)), msd[:, 0], msd[:, 1])]
        figset = {'xlabel': 'Time', 'ylabel': 'MSD'}
        canvas[ORDERFILES['msd']] = mpl_error_plot(series,
                                                   fig_settings=figset)
    return canvas


def mpl_plot_energy(results, energies):
    """Plot the output from the energy analysis using matplotlib.

    Parameters
    ----------
    results : dict
        Each item in `results` contains the results for the
        corresponding energy. It is assumed to contains the keys
        'vpot', 'ekin', 'etot', 'ham', 'temp', 'elec'.
    energies : dict of numpy.arrays
        This is the raw data for the energy analysis.

    Returns
    -------
    canvas : dict
        The output figures created by this function.

    """
    canvas = {}
    time = energies['time']
    # make time series plot of the energies
    series = []
    for key in ['vpot', 'ekin', 'etot', 'ham']:
        if key not in energies:
            continue
        series.append({'type': 'xy', 'x': time, 'y': energies[key],
                       'label': ENERTITLE[key]})
    figset = {'xlabel': 'Time', 'ylabel': 'Energy'}
    canvas[ENERFILES['energies']] = mpl_simple_plot(series,
                                                    fig_settings=figset)
    # make running average plot of the energies as function of time
    series = []
    for key in ['vpot', 'ekin', 'etot', 'ham']:
        if key not in results:
            continue
        series.append({'type': 'xy', 'x': time,
                       'y': results[key]['running'],
                       'label': ENERTITLE[key]})
    canvas[ENERFILES['run_energies']] = mpl_simple_plot(series,
                                                        fig_settings=figset)
    # plot temperature
    series = [{'type': 'xy', 'x': time, 'y': energies['temp']}]
    figset = {'xlabel': 'Time', 'ylabel': 'Temperature'}
    canvas[ENERFILES['temperature']] = mpl_simple_plot(series,
                                                       fig_settings=figset)
    # and running average for temperature
    series = [{'type': 'xy', 'x': time, 'y': results['temp']['running']}]
    figset = {'xlabel': 'Time',
              'ylabel': 'Temperature',
              'title': 'Running average'}
    canvas[ENERFILES['run_temp']] = mpl_simple_plot(series,
                                                    fig_settings=figset)
    # plot block-error results:
    title = r'{0}: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
    for key in ['vpot', 'ekin', 'etot', 'temp']:
        if key not in results:
            continue
        plot = ENERFILES['block'].format(key)
        block = results[key]['blockerror']
        series = [
            {'type': 'xy', 'x': block[0], 'y': block[3]},
            {'type': 'hline', 'y': block[4], 'ls': '--', 'alpha': 0.8}
        ]
        figset = {'xlabel': 'Block length',
                  'ylabel': 'Estimated error',
                  'title': title.format(ENERTITLE[key], block[4], block[6])}
        canvas[plot] = mpl_simple_plot(series, fig_settings=figset)
    # plot distributions
    for key in ['vpot', 'ekin', 'etot', 'temp']:
        if key not in results:
            continue
        dist = results[key]['distribution']
        series = [{'type': 'xy', 'x': dist[1], 'y': dist[0],
                   'label': ENERTITLE[key]}]
        title = '{0}. Average: {1:9.6e}, std: {2:9.6f}'
        title = title.format(ENERTITLE[key], dist[2][0], dist[2][1])
        if 'boltzmann-dist' in results[key]:
            series.append({'type': 'xy',
                           'x': results[key]['boltzmann-dist'][1],
                           'y': results[key]['boltzmann-dist'][0],
                           'label': 'Boltzmann distribution'})
        plot = ENERFILES['dist'].format(key)
        canvas[plot] = mpl_simple_plot(series, fig_settings={'title': title})
    return canvas


def mpl_plot_flux(results):
    """Plot the output from the flux analysis using matplotlib.

    Parameters
    ----------
    results : dict
        This is the dict with the results from the flux analysis.

    Returns
    -------
    out[0] : list of dicts
        The output figures created by this function for running
        averages. `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.
    out[1] : list of dicts
        The output figures created by this function for block errors.
        `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.

    """
    canvas_run = []
    canvas_err = []
    for i in range(len(results['flux'])):
        # Plot running average:
        flux = results['flux'][i]
        runflux = results['runflux'][i]
        series = [{'type': 'xy', 'x': flux[:, 0], 'y': runflux,
                   'label': 'Running average'}]
        title = f'Flux for interface no. {i + 1}'
        figset = {'xlabel': 'Time',
                  'ylabel': 'Flux / internal units',
                  'title': title}
        canvas = mpl_simple_plot(series, fig_settings=figset)
        canvas_run.append({'name': FLUXFILES['runflux'].format(i + 1),
                           'canvas': canvas})
        # Plot error results:
        errflux = results['errflux'][i]
        title = r'Block error: {0}: Rel. err.: {1:9.6e}, Ncor: {2:9.6f}'
        series = [{'type': 'xy', 'x': errflux[0], 'y': errflux[3]},
                  {'type': 'hline', 'y': errflux[4], 'ls': '--',
                   'alpha': 0.8}]
        figset = {'xlabel': 'Block length',
                  'ylabel': 'Estimated error',
                  'title': title.format(i + 1, errflux[4], errflux[6])}
        canvas = mpl_simple_plot(series, fig_settings=figset)
        canvas_err.append({'name': FLUXFILES['block'].format(i + 1),
                           'canvas': canvas})
    return canvas_run, canvas_err


def mpl_plot_xi(xi, xierror):
    """Plot the output from the Xi analysis using matplotlib.

    Parameters
    ----------
    xi: array-like
        The running average of xi
    xierror: tuple
        The results from the blockerror analysis of xi

    Returns
    -------
    out[0] : list of dicts
        The output figures created by this function for running
        averages. `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.
    out[1] : list of dicts
        The output figures created by this function for block errors.
        `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.

    """
    canvas_run = []
    canvas_err = []
    series = [{'type': 'xy', 'x': range(len(xi)), 'y': xi,
               'label': 'Running average'},
              {'type': 'hline', 'y': xi[-1], 'ls': '--', 'alpha': 0.8}]
    title = r'$\xi={}$:'
    figset = {'xlabel': 'cycles',
              'ylabel': r'$\xi$',
              'title': title.format(xi[-1])}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    canvas_run.append({'name': "xirun",
                       'canvas': canvas})
    # Plot error results:
    title = r"relative $\xi$ error: Rel. err.: {0:9.6e}, Ncor: {1:9.6f}"
    series = [{'type': 'xy', 'x': xierror[0], 'y': xierror[3]},
              {'type': 'hline', 'y': xierror[4], 'ls': '--',
               'alpha': 0.8}]
    figset = {'xlabel': 'Block length',
              'ylabel': 'Estimated error',
              'title': title.format(xierror[4], xierror[6])}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    canvas_err.append({'name': "xierr",
                       'canvas': canvas})

    return canvas_run, canvas_err


def mpl_plot_tau(tau, tauerror):
    """Plot the output from the Tau analysis using matplotlib.

    Parameters
    ----------
    tau: array-like
        The running average of tau.
    tauerror: tuple
        The results from the blockerror analysis of tau.

    Returns
    -------
    out[0] : list of dicts
        The output figures created by this function for running
        averages. `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.
    out[1] : list of dicts
        The output figures created by this function for block errors.
        `out[0][i]['name']` is the name of the figure and
        `out[0][i]['canvas']` is the corresponding canvas object.

    """
    canvas_run = []
    canvas_err = []
    if not tau.any():
        str_in = "no"
    else:
        str_in = "in"
    series = [{'type': 'xy', 'x': range(len(tau)), 'y': tau,
               'label': 'Running average'},
              {'type': 'hline', 'y': tau[-1], 'ls': '--', 'alpha': 0.8}]
    title = (r'$\frac{\tau}{dz}$ '
             f"({str_in} "
             r'selected $\tau$ region)$= $')
    figset = {'xlabel': 'cycles',
              'ylabel': r'$\frac{\tau}{dz}$',
              'title': title+f"${tau[-1]}$"}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    canvas_run.append({'name': "taurun",
                       'canvas': canvas})
    # Plot error results:
    title = (r"relative error $\frac{\tau}{dz}$ "
             f"({str_in} "
             r"selected $\tau$ region):")
    title2 = r"Rel. err.: {0:9.6e}, Ncor: {1:9.6f}"
    series = [{'type': 'xy', 'x': tauerror[0], 'y': tauerror[3]},
              {'type': 'hline', 'y': tauerror[4], 'ls': '--',
               'alpha': 0.8,
               'label': title2.format(tauerror[4], tauerror[6])}]
    figset = {'xlabel': 'Block length',
              'ylabel': 'Estimated error',
              'title': title}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    canvas_err.append({'name': "tauerr",
                       'canvas': canvas})

    return canvas_run, canvas_err


def mpl_plot_tau_ref(tau_ref, tau_ref_bins, bins):
    """Plot a histogram of the Tau reference bins using matplotlib.

    Parameters
    ----------
    tau_ref: array-like
        The final running average of all tau ref bins.
    tau_ref_bins: list of tuple
        The (L,R) bin edges of the reference bins.
    bins: tuple
        The (L,R) bin edges of the region used to calculate tau.

    Returns
    -------
    out : dict
        The output figure created by this function for histograming the
        tau reference bins. dict['name']` is the name of the figure and
        `dict['canvas']` is the corresponding canvas object.

    """
    x = [i[0] for i in tau_ref_bins]
    width = x[1]-x[0]
    series = [{'type': 'bar', 'x': x, 'height': tau_ref,
               'width': width, 'align': 'edge'}]
    if len(bins) == 2:
        series.append(
              {'type': 'vline', 'x': bins[0], 'ls': '--', 'alpha': 0.8,
               'label': r'selected $\tau$ region'}
        )
        series.append(
              {'type': 'vline', 'x': bins[1], 'ls': '--', 'alpha': 0.8}
        )
    title = r'$\frac{\tau}{dz}$ reference bins'
    figset = {'xlabel': 'Orderparameter',
              'ylabel': r'$\frac{\tau}{dz}$ count',
              'title': title}
    canvas = mpl_simple_plot(series, fig_settings=figset)
    return {"name": "tauref", "canvas": canvas}


def get_color_map(ncolors):
    """Return a color map with at least n colors."""
    if ncolors <= 10:
        name = 'tab10'
    elif 10 < ncolors <= 20:
        name = 'tab20'
    else:
        name = None
    if name is None:
        logger.info('Using default color map.')
    else:
        logger.info('Using color map %s', name)
    cmap = matplotlib.colormaps[name]
    return cmap(np.linspace(0, 1, ncolors))


def mpl_plot_matched(path_ensembles, detect, matched, reptis=False):
    """Plot matched probabilities using matplotlib.

    This function will plot the matched probabilities for the different
    ensembles and also make a plot with just the over-all matched
    probability.

    Parameters
    ----------
    path_ensembles : list of strings
        This is the name of the path ensembles we have calculated
        the probability for.
    detect : list of floats
        These are the detect interfaces used in the analysis.
    matched : dict
        This dict contains the results from the matching of the
        probabilities: `matched['overall-prob']`,
         and `matched['overall-prun']`
         or `matched['overall-rrun']` (if exists) are used here.

    Returns
    -------
    canvas : dict
        The output figures created by this function.

    """
    canvas = {}
    # Also make a plot with the overall matched probability:
    series = []
    for idetect in detect:
        series.append({'type': 'vline', 'x': idetect,
                       'ls': '--', 'alpha': 0.8, 'lw': 1})
    if reptis:
        matched['overall-prob'] = np.array([detect,
                                            matched['pcross']]).T
    series.append({'type': 'xy',
                   'x': matched['overall-prob'][:, 0],
                   'y': matched['overall-prob'][:, 1],
                   'lw': 3})
    figset = {'xlabel': r'Order parameter ($\lambda$)',
              'ylabel': 'Crossing probability',
              'title': 'Matched probability',
              'yscale': 'log'}
    canvas[PATH_MATCH['match']] = mpl_simple_plot(series,
                                                  fig_settings=figset)
    # Also make a plot with the TIME evolution of the overall
    # probability or rate:
    if 'overall-rrun' in matched:
        title = 'Time evolution of the overall Rate'
        ylabel = 'Rate (running estimate)'
        data = matched['overall-rrun']
        label = 'overall-rrun'

    elif 'overall-prun' in matched:
        title = 'Time evolution of the overall crossing probability'
        ylabel = 'Crossing probability (running estimate)'
        data = matched['overall-prun']
        label = 'overall-prun'
    else:
        return canvas

    series = [{'type': 'xy',
               'x': matched['overall-cycle'],
               'y': data},
              {'type': 'hline',
               'y': data[-1],
               'ls': '--',
               'alpha': 0.8}]
    figset = {'xlabel': 'Cycle number',
              'ylabel': ylabel,
              'title': title}
    canvas[label] = mpl_simple_plot(series, fig_settings=figset)
    # and finish with the block error analysis.
    series = [{'type': 'xy',
               'x': matched['overall-error'][0],
               'y': matched['overall-error'][3]},
              {'type': 'hline',
               'y': matched['overall-error'][4],
               'ls': '--',
               'alpha': 0.8}]
    if reptis:
        title = 'Relative error of the running crossing probability'
    else:
        title = r'Overall: Rate Rel. err.: {0:9.6e}, Ncor: {1:9.6e}'
        title.format(matched['overall-error'][4], matched['overall-error'][6])
    figset = {'xlabel': 'Block length', 'ylabel': 'Estimated error',
              'title': title}
    canvas[PATH_MATCH['error']] = mpl_simple_plot(series,
                                                  fig_settings=figset)
    if reptis:
        # reptis does not deal with ..
        return canvas

    # First plot the matched probabilities for each ensemble:
    series = []
    for idetect in detect:
        series.append({'type': 'vline', 'x': idetect,
                       'ls': '--', 'alpha': 0.8, 'lw': 1})

    # color_cycle was deprecated in matplotlib 1.5, but to support old
    # versions:
    if 'axes.prop_cycle' in matplotlib.rcParams:
        ckey = 'axes.prop_cycle'
    else:
        ckey = 'axes.color_cycle'
    # Check if we need to have more colors:
    if len(matplotlib.rcParams[ckey]) < len(path_ensembles):
        logger.info('Overriding color cycle.')
        colors = get_color_map(len(path_ensembles))
    else:
        colors = None

    new_series = {'type': 'xy',
                  'x': matched['overall-prob'][:, 0],
                  'y': matched['overall-prob'][:, 1],
                  'alpha': 0.8,
                  'lw': 9, 'label': 'Overall', 'color': '#262626'}
    series.append(new_series)
    for i, (prob, path_e) in enumerate(zip(matched['matched-prob'],
                                           path_ensembles)):
        new_series = {'type': 'xy', 'x': prob[:, 0], 'y': prob[:, 1], 'lw': 3,
                      'label': f'${path_e}$'}
        if colors is not None:
            new_series['color'] = colors[i]
        series.append(new_series)

    figset = {'xlabel': r'Order parameter ($\lambda$)',
              'ylabel': 'Probability',
              'title': 'Overall probabilities',
              'yscale': 'log'}
    canvas[PATH_MATCH['total']] = mpl_simple_plot(series,
                                                  fig_settings=figset)
    return canvas


def mpl_plot_bar(axs, seri):
    del seri['type']
    return axs.bar(**seri)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of interaction with a Path."""
import numpy as np
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.core.path import Path
from pyretis.core.random_gen import RandomGenerator

path = Path(rgen=RandomGenerator(seed=0))  # Create empty path.

# Add some phase points to the path:
for i in range(10):
    phasepoint = System()
    phasepoint.order = [i]
    phasepoint.particles = Particles(dim=3)
    phasepoint.add_particle(np.zeros(3), vel=np.zeros(3))
    phasepoint.vpot = i
    phasepoint.ekin = i
    path.append(phasepoint)

# Loop over the phase points in the path:
print('Looping forward:')
for i, phasepoint in enumerate(path.phasepoints):
    print('Point {}. Order parameter = {}'.format(i, phasepoint.order))

# Loop over the phase points in the path:
print('Looping backward:')
for phasepoint in reversed(path.phasepoints):
    print('Order parameter = {}'.format(phasepoint.order))

# Get some randomly chosen shooting points:
print('Generating shooting points:')
for i in range(10):
    point, idx = path.get_shooting_point()
    print(
        'Shooting point {}: index = {}, order = {}'.format(i, idx, point.order)
    )



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for formatting path ensemble data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PathEnsembleFormatter (:py:class:`.PathEnsembleFormatter`)
    A class for formatting path ensemble data.

PathEnsembleFile (:py:class:`.PathEnsembleFile`)
    A class for handling PyRETIS path ensemble files.

"""
import logging
from pyretis.core.pathensemble import PathEnsemble
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.fileio import FileIO
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['PathEnsembleFormatter', 'PathEnsembleFile']


class PathEnsembleFormatter(OutputFormatter):
    """A class for formatting path ensemble data.

    This class will effectively define the data which we store
    for each path ensemble. The data is stored in columns with the
    format defined below and contains:

    1) The current cycle number: `'Step'`.

    2) The number of accepted paths: `'No.-acc'`.

    3) The number of shooting moves attempted: `'No.shoot'`.

    4) Starting location (with respect to interfaces): `'l'`.
       This can be `L` if a path starts on the left side, `R` if it
       starts on the right side and `*` if it did not reach the
       interface.

    5) Marker for crossing he middle interface: (`'m'`). This is
       either `M` (when the middle interface is crossed by the path)
       or `*` (if the middle interface is no crossed).

    6) End point for the path: `'r'`. The possible values here are
       similar to the values for the starting location given above.

    7) The length of the generated path: `'Length'`.

    8) The status of the path: `'Acc'`. This is one of the possible
       path statuses defined in :py:mod:`pyretis.core.path`.

    9) The type of move executed for generating the path: `'Mc'`.
       This is one of the moves defined in :py:mod:`pyretis.core.path`.

    10) The smallest order parameter in the path: `'Min-O'`.

    11) The largest order parameter in the path: `'Max-O'`.

    12) The index in the path for the smallest order parameter: `'Idx-Min'`.

    13) The index in the path for the largest order parameter: `'Idx-Max'`.

    14) The order parameter for the shooting point, immediately after
        the shooting move: `'O-shoot'`.

    15) The index in the source path for the shooting point: `'Idx-sh'`.

    16) The index in the new path for the shooting point: `'Idx-shN'`.

    17) The statistical weight of the path: `'Weight'`.

    """

    PATH_FMT = (
        '{0:>10d} {1:>10d} {2:>10d} {3:1s} {4:1s} {5:1s} {6:>7d} '
        '{7:3s} {8:2s} {9:>16.9e} {10:>16.9e} {11:>7d} {12:>7d} '
        '{13:>16.9e} {14:>7d} {15:7d} {16:>16.9e}'
    )
    HEADER = {
        'labels': ['Step', 'No.-acc', 'No.-shoot',
                   'l', 'm', 'r', 'Length', 'Acc', 'Mc',
                   'Min-O', 'Max-O', 'Idx-Min', 'Idx-Max',
                   'O-shoot', 'Idx-sh', 'Idx-shN', 'Weight'],
        'width': [10, 10, 10, 1, 1, 1, 7, 3, 2, 16, 16, 7, 7, 16, 7, 7, 7],
    }

    def __init__(self):
        """Initialise the formatter for path ensemble data."""
        super().__init__('PathEnsembleFormatter', header=self.HEADER)
        self.print_header = True

    def format(self, step, data):
        """Format path ensemble data.

        Here we generate output based on the last path from the
        given path ensemble.

        Parameters
        ----------
        step : int
            The current cycle/step number.
        data : object like :py:class:`.PathEnsemble`
            The path ensemble we will format here.

        Yields
        ------
        out : string
            The formatted line with the data from the last path.

        """
        path_ensemble = data
        path_dict = path_ensemble.paths[-1]
        interfaces = ['*' if i is None else i for i in path_dict['interface']]

        yield self.PATH_FMT.format(
            step,
            path_ensemble.nstats['ACC'],
            path_ensemble.nstats['nshoot'],
            interfaces[0],
            interfaces[1],
            interfaces[2],
            path_dict['length'],
            path_dict['status'],
            path_dict['generated'][0],
            path_dict['ordermin'][0],
            path_dict['ordermax'][0],
            path_dict['ordermin'][1],
            path_dict['ordermax'][1],
            path_dict['generated'][1],
            path_dict['generated'][2],
            path_dict['generated'][3],
            path_dict['weight'],
        )

    def parse(self, line):
        """Parse a line to a simplified representation of a path.

        Parameters
        ----------
        line : string
            The line of text to parse.

        Returns
        -------
        out : dict
            The dict with the simplified path information.

        """
        linec = line.strip()
        if linec.startswith('#'):
            # This is probably the comment
            return None
        # stip trailing comments
        linec = linec.split('#')[0]
        data = [i.strip() for i in linec.split()]
        if len(data) < 16:
            logger.warning(
                'Incorrect number of columns in path data, skipping line.'
            )
            return None
        if len(data) == 16:
            path_info = {'cycle': int(data[0]),
                         'generated': [str(data[8]), float(data[13]),
                                       int(data[14]), int(data[15])],
                         'interface': (str(data[3]), str(data[4]),
                                       str(data[5])),
                         'length': int(data[6]),
                         'ordermax': (float(data[10]), int(data[12])),
                         'ordermin': (float(data[9]), int(data[11])),
                         'status': str(data[7]),
                         'weight': 1.}

        else:
            path_info = {
                'cycle': int(data[0]),
                'generated': [str(data[8]), float(data[13]),
                              int(data[14]), int(data[15])],
                'interface': (str(data[3]), str(data[4]), str(data[5])),
                'length': int(data[6]),
                'ordermax': (float(data[10]), int(data[12])),
                'ordermin': (float(data[9]), int(data[11])),
                'status': str(data[7]),
                'weight': float(data[16]),
            }

        return path_info

    def load(self, filename):
        """Yield the different paths stored in the file.

        The lines are read on-the-fly, converted and yielded one-by-one.

        Parameters
        ----------
        filename : string
            The path/filename to open.

        Yields
        ------
        out : dict
            The information for the current path.

        """
        try:
            with open(filename, 'r', encoding="utf8") as fileh:
                for line in fileh:
                    path_data = self.parse(line)
                    if path_data is not None:
                        path_data['filename'] = filename
                        yield path_data
        except IOError as error:
            logger.critical('I/O error (%d): %s', error.errno, error.strerror)
        except Exception as error:  # pragma: no cover
            logger.critical('Error: %s', error)
            raise


class PathEnsembleFile(PathEnsemble, FileIO):
    """A class for handling PyRETIS path ensemble files.

    This class inherits from the :py:class:`.PathEnsemble` and
    this makes it possible to run the analysis directly on this
    file.

    """

    def __init__(self, filename, file_mode, ensemble_settings=None,
                 backup=True):
        """Set up the file-like object.

        Parameters
        ----------
        filename : string
            The file to open.
        file_mode : string
            Determines the mode for opening the file.
        ensemble_settings : dict, optional
            Ensemble specific settings.
        backup : boolean, optional
            Determines how we handle existing files when the mode
            is set to writing.

        """
        default_settings = {
            'ensemble_number': 0,
            'interfaces': [0.0, 1.0, 2.0],
            'detect': None
        }
        settings = {}
        for key, val in default_settings.items():
            try:
                settings[key] = ensemble_settings[key]
            except (TypeError, KeyError):
                settings[key] = val
                logger.warning(
                    'No "%s" ensemble setting given for "%s". Using defaults',
                    key,
                    self.__class__
                )
        PathEnsemble.__init__(self, settings['ensemble_number'],
                              settings['interfaces'])
        FileIO.__init__(self, filename, file_mode, PathEnsembleFormatter(),
                        backup=backup)

    def get_paths(self):
        """Yield paths from the file."""
        return self.load()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Helper to create tables for PyRETIS keywords."""
import textwrap
import pathlib


def make_table(data, short_refs):
    """Make the actual table."""
    header1 = 'Keyword'
    maxlength = 79 - 3
    length = [len(i.split('|')[1]) + 2 for i in short_refs]
    length.append(len(header1))
    length1 = max(length)
    fmt1 = '| {{:<{}s}} |'.format(length1)
    border1 = '+' + '-' * (length1 + 2) + '+'
    eborder1 = '+' + '=' * (length1 + 2) + '+'
    rem = maxlength - len(border1)
    length2 = maxlength - length1 - 4 - 3
    fmt2 = ' {{:<{}s}} |'.format(length2)
    border2 = '-' * (rem - 1) + '+'
    eborder2 = '=' * (rem - 1) + '+'
    table = []
    table.append(border1 + border2)
    table.append(fmt1.format(header1) + fmt2.format('Description'))
    table.append(eborder1 + eborder2)

    for refi, datai in zip(short_refs, data):
        try:
            txt = (' '.join(datai)).split('>`:')[1].strip()
            txtl = textwrap.wrap(txt, width=length2)
        except IndexError:
            print('NO DATA FOUND!')
            txtl = ['']
        col1 = []
        col2 = []
        for i, txt in enumerate(txtl):
            if i == 0:
                col1.append(fmt1.format(refi.split()[1]))
            else:
                col1.append(fmt1.format(''))
            col2.append(fmt2.format(txt))
        for i, j in zip(col1, col2):
            table.append(i + j)
        table.append(border1 + border2)

    for ref in short_refs:
        print('\n{}'.format(ref))

    print()
    print('.. _table-NAMEOFTABLE:')
    print()
    print('.. table:: Supported XXX in |pyretis|')
    print('   :class: table-striped')
    print()

    for line in table:
        print('   {}'.format(line))


def shorten_ref(ref, add_to_ref):
    """Shorten references."""
    link = ':ref:{}'.format(ref.split(':ref:')[1].split(':')[0])
    linkr = link.split('<')[1].split('>')[0]
    short = '.. |{}_{}| replace:: {}'.format(add_to_ref,
                                             linkr.split('-')[-1],
                                             link)
    print(short.split('|')[1])
    return short


def read_keyword_section(filename, add_to_ref):
    """Read a keyword section in a .rst file."""
    read = False
    data = []
    sections = []
    section = []
    new_section = True
    filename = pathlib.Path(filename)
    with filename.open() as fileobj:
        for lines in fileobj:
            if lines.startswith('* :ref:'):
                if new_section:
                    if section:
                        sections.append(section)
                    section = []
                    new_section = False
                if data:
                    section.append(data)
                data = []
                read = True
            else:
                if lines.startswith('..'):
                    new_section = True
            if not lines.strip():
                read = False
            if read:
                data.append(lines.strip())
        if data:
            section.append(data)
        if section:
            sections.append(section)

    for section in sections:
        short = [shorten_ref(data[0], add_to_ref) for data in section]
        make_table(section, short)


if __name__ == '__main__':
    read_keyword_section('unitsystem.rst', 'units')



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Dummy potential for tests."""
from pyretis.forcefield.potential import PotentialFunction


__all__ = ['FooPotential']


class FooPotential(PotentialFunction):
    """FooPotential(PotentialFunction) - Dummy potential for tests."""

    def __init__(self, a=0.0, desc='Dummy potential'):
        super().__init__(dim=1, desc=desc)
        self.params = {'a': a}

    def potential(self, pos):
        """Evaluate the potential."""
        raise NotImplementedError

    def force(self, pos):
        """Evaluate the force."""
        raise NotImplementedError

    def potential_and_force(self, pos):
        """Evaluate the potential and force."""
        raise NotImplementedError


class BarPotential(PotentialFunction):
    """BarPotential(PotentialFunction) - Dummy potential for tests."""

    def __init__(self, a=0.0, desc='Dummy potential'):
        super().__init__(dim=1, desc=desc)
        self.params = {'a': a}

    def potential(self, pos):
        """Evaluate the potential."""
        raise NotImplementedError

    def force(self, pos):
        """Evaluate the force."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The Langevin integrator implemented in FORTRAN."""
import logging
import os
import sys
import numpy as np
from pyretis.engines import MDEngine
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
try:
    from vvintegrator import vvintegrator
except ImportError:
    MSG = ('Could not import external FORTRAN library.'
           '\nPlease compile with "make"!')
    logger.critical(MSG)
    raise ImportError(MSG)


__all__ = ['LangevinF']


class LangevinF(MDEngine):
    """LangevinF(MDEngine).

    This class defines the Langevin integrator.

    Attributes
    ----------
    timestep : float
        The time step.
    gamma : float
        The gamma parameter for the Langevin integrator.
    high_friction : boolean
        Determines if we are in the high_friction limit and should
        do the over-damped version.

    """

    def __init__(self, timestep, gamma, high_friction, seed):
        """Set up the Velocity Verlet integrator.

        Parameters
        ----------
        timestep : float
            The time step.
        gamma : float
            The gamma parameter for the Langevin thermostat
        high_friction : boolean
            Determines if we should do overdamped or not.
        seed : integer
            A seed for the random number generator.

        """
        super().__init__(timestep, 'The Langevin integrator (FORTRAN)',
                         dynamics='stochastic')
        self.gamma = gamma
        self.high_friction = high_friction
        self.param_high = {'sigma': None, 'bddt': None}
        self.param_iner = {'c0': None, 'a1': None, 'a2': None, 'b1': None,
                           'b2': None, 's12os11': None, 'sqrts11': None,
                           'sqrtsos11': None}
        self.init_params = True
        size = vvintegrator.get_seed_size()
        seeds = np.array([seed + i for i in range(size)], dtype='i4')
        vvintegrator.seed_random_generator(seeds)

    def _initiate_parameters(self, system):
        """Set up Langevin parameters.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in `system.particles`.

        """
        beta = system.temperature['beta']
        imasses = system.particles.imass
        if self.high_friction:
            self.param_high['sigma'] = np.sqrt(2.0 * self.timestep *
                                               imasses/(beta * self.gamma))
            self.param_high['bddt'] = self.timestep * imasses / self.gamma
        else:
            gammadt = self.gamma * self.timestep
            exp_gdt = np.exp(-gammadt)
            s11 = ((self.timestep * imasses / (beta * self.gamma)) *
                   (2. - (3. - 4. * exp_gdt + exp_gdt**2) / gammadt))
            sig_r = np.sqrt(s11)
            s22 = (1.0 - exp_gdt**2) * imasses / beta
            s12 = (imasses / (beta * self.gamma)) * (1. - exp_gdt)**2
            self.param_iner['s12os11'] = s12 / s11
            self.param_iner['sqrts11'] = sig_r
            self.param_iner['sqrtsos11'] = np.sqrt((s11 * s22 - s12**2) / s11)
            if self.gamma > 0.0:
                c_0 = exp_gdt
                c_1 = (1.0 - c_0) / gammadt
                c_2 = (1.0 - c_1) / gammadt
            else:
                c_0, c_1, c_2 = 1.0, 1.0, 0.5
            self.param_iner['c0'] = c_0
            self.param_iner['a1'] = c_1 * self.timestep
            self.param_iner['a2'] = c_2 * self.timestep**2 * imasses
            self.param_iner['b1'] = (c_1 - c_2) * self.timestep * imasses
            self.param_iner['b2'] = c_2 * self.timestep * imasses

    def integration_step(self, system):
        """Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in `system.particles`.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        if self.init_params:
            self._initiate_parameters(system)
            self.init_params = False
        if self.high_friction:
            return self.integration_step_overdamped(system)
        return self.integration_step_inertia(system)

    def integration_step_overdamped(self, system):
        """Over damped Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        system.force()  # update forces
        particles = system.particles
        particles.pos, particles.vel = vvintegrator.overdamped(
            particles.pos,
            particles.vel,
            particles.force,
            self.param_high['bddt'],
            self.param_high['sigma']
        )
        system.potential()

    def integration_step_inertia(self, system):
        """Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        particles.pos, particles.vel = vvintegrator.inertia1(
            particles.pos,
            particles.vel,
            particles.force,
            self.gamma,
            self.param_iner['c0'],
            self.param_iner['a1'],
            self.param_iner['a2'],
            self.param_iner['b1'],
            self.param_iner['s12os11'],
            self.param_iner['sqrts11'],
            self.param_iner['sqrtsos11']
        )
        system.force()  # update forces
        particles.vel = vvintegrator.inertia2(
            particles.vel,
            particles.force,
            self.param_iner['b2']
        )
        system.potential()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is a mock engine for external testing.

This script is inteneded to be used as an external engine and will
create some LAMMPS-like output.
"""
import sys
from pyretis.engines.lammps import read_lammps_input


def get_arguments(lammps_args):
    """Read arguments given to this program.

    Note, we here assuming that the arguments we are served
    are on the form: -in <inputfile> -l <logfile> -screen <screenfile>

    """
    looking_for = {'-in', '-l', '-screen'}
    found = {}
    for i, arg in enumerate(lammps_args):
        if arg in looking_for:
            var = lammps_args[i+1]
            found[arg] = var
    errors = False
    for key in looking_for:
        if key not in found:
            errors = True
            print(f'ERROR: Missing input argument "{key}"')
    if errors:
        sys.exit(1)
    return found


def write_fake_log(logfile, steps, freq):
    """Create a fake log file for LAMMPS."""
    with open(logfile, 'w', encoding='utf-8') as output:
        output.write('This is not LAMMPS.\n')
        output.write('But we are pretending to be.\n\n')
        output.write('Please have some fake thermo data:\n')
        output.write('Step Temp Press PotEng KinEng TotEng\n')
        for i in range(steps + 1):
            if i % freq == 0:
                j = 5 * i * freq
                output.write(
                    f'{i} {j + 1} {j + 2} {j + 3} {j + 4} {j + 5}\n'
                )
        output.write('Loop time of 0 on 1 procs for 12 steps with 0 atoms\n')
        output.write('\n')
        output.write('Total wall time: 0:00:00')


def write_fake_screen(screenfile, settings):
    """Create a fake output file for LAMMPS.

    Here, we just repeat the input settings.

    """
    with open(screenfile, 'w', encoding='utf-8') as output:
        for key, val in settings:
            output.write(f'{key} {val}\n')


def make_dump_file(filename, steps, freq):
    """Create a fake dump file."""
    with open(filename, 'w', encoding='utf-8') as output:
        output.write('# Fake dump file.')
        for i in range(steps + 1):
            if i % freq == 0:
                output.write('ITEM: TIMESTEP\n')
                output.write(f'{i}\n')


def make_restart_file(filename):
    """Create a fake restart file."""
    with open(filename, 'w', encoding='utf-8') as output:
        output.write('Fake restart file.')


def make_order_file(filename, steps, freq):
    """Create a fake order parameter file."""
    with open(filename, 'w', encoding='utf-8') as output:
        output.write('# Fake order parameter file.\n')
        for i in range(steps + 1):
            if i % freq == 0:
                output.write(
                    f'{i} {i + 1} {-(i + 1)}\n'
                )


def make_fake_output_files(settings, steps):
    """Create output files as requested by the settings."""
    for key, val in settings:
        if key == 'dump':
            # Assume that the setting is: dump ID group-ID style N filename
            filename = val.split()[4].strip()
            freq = int(val.split()[3].strip())
            print(f'Dumping to: "{filename}"')
            make_dump_file(filename, steps, freq)
        elif key == 'write_restart':
            filename = val.split()[0].strip()
            print(f'Writing restart to: "{filename}"')
            make_restart_file(filename)
        elif key == 'fix':
            # Check if we are to write order parameters.
            if val.startswith('order_output'):
                val_split = [i.strip() for i in val.split()]
                idx = val_split.index('append')
                filename = val_split[idx+1]
                idx = val_split.index('print')
                freq = int(val_split[idx+1])
                print(f'Writing order parameters to: "{filename}"')
                make_order_file(filename, steps, freq)


def main(lammps_args):
    """Run the mock LAMMPS engine."""
    arguments = get_arguments(lammps_args)
    settings = read_lammps_input(arguments['-in'])
    # First just figure out the number of steps:
    steps = 0
    freq = 1
    for key, val in settings:
        if key == 'run' and len(val.split()) == 1:
            steps = int(val)
        if key == 'thermo' and len(val.split()) == 1:
            freq = int(val)
    # Create the output files
    write_fake_log(arguments['-l'], steps, freq)
    write_fake_screen(arguments['-screen'], settings)
    make_fake_output_files(settings, steps)


if __name__ == '__main__':
    print('This is the mock LAMMPS command')
    main(sys.argv)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test methods for doing TIS."""
import copy
import logging
import os
import tempfile
import unittest
from unittest.mock import patch
from io import StringIO
import numpy as np
from pyretis.core import System, create_box, Particles
from pyretis.core.common import big_fat_comparer
from pyretis.core.pathensemble import PathEnsemble
from pyretis.core.random_gen import (
    MockRandomGenerator,
    RandomGenerator,
    create_random_generator)
from pyretis.core.tis import (
    extender,
    make_tis,
    make_tis_step_ensemble,
    paste_paths,
    select_shoot,
    shoot,
    stone_skipping,
    wire_fencing,
    ss_wt_wf_metropolis_acc,
    ss_wt_wf_acceptance,
    time_reversal,
    web_throwing,
)
from pyretis.forcefield import ForceField
from pyretis.engines.internal import VelocityVerlet, RandomWalk
from pyretis.initiation import initiate_path_simulation
from pyretis.inout.checker import check_engine
from pyretis.inout.checker import check_ensemble
from pyretis.inout.common import make_dirs
from pyretis.inout.settings import fill_up_tis_and_retis_settings
from pyretis.orderparameter import Position
from pyretis.setup import create_force_field, create_engine
from pyretis.setup.createsimulation import (create_ensemble,
                                            create_tis_simulation)
from .help import (
    create_ensembles_and_paths,
    make_internal_path,
    MockEngine,
    MockEngine2,
    MockOrder,
    MockOrder2,
    MockPotential,
    NegativeOrder,
    SameOrder,
    prepare_test_settings,
    prepare_test_simulation
)

logging.disable(logging.CRITICAL)

TISMOL_001 = [[262, 'ACC', 'ki', -0.903862, 1.007330, 1, 262],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [12, 'BWI', 'sh', 0.957091, 1.002750, 12, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [77, 'BWI', 'sh', 0.522609, 1.003052, 77, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1],
              [262, 'BWI', 'tr', -0.903862, 1.007330, 262, 1]]


def create_ensemble_and_paths(example=0):
    """Return some test data we can use."""
    # Make some paths for the ensembles.
    interfaces = [(-1., -1., 10.), (-1., 0., 10.),
                  (-1., 1., 10.), (-1., 2., 10.)]
    starts = [(0, -1.1), (0, -1.05), (0, -1.123), (0, -1.01)]
    ends = [(100, -1.2), (100, -1.3), (100, -1.01),
            (100, 10.01)]
    maxs = [(50, -0.2), (50, 0.5), (50, 2.5), (100, 10.01)]

    settings = prepare_test_settings()
    start = starts[example]
    end = ends[example]
    mxx = maxs[example]
    settings['simulation']['interfaces'] = interfaces[example]
    settings['simulation']['zero_ensemble'] = False
    if example == 3:
        settings['tis']['detect'] = interfaces[example][2]
    else:
        settings['tis']['detect'] = interfaces[example][1]
    settings['tis']['ensemble_number'] = example
    with patch('sys.stdout', new=StringIO()):
        ensemble = create_ensemble(settings)
    path = make_internal_path(start, end, mxx, ensemble['interfaces'][1])
    ensemble['path_ensemble'].add_path_data(path, 'ACC')
    return ensemble


def create_ensemble_and_paths_repptis(example=0):
    """Creating dummy ensemble for repptis shooting tests.
    We have 6 ensembles, with the following acceptable paths:
        [0^-]---RMR
        [0^+-']-LML
        [1^+-]--LMR
        [2^+-]--LML
        [3^+-]--RMR
        [4^+-]--LML
    """
    interfaces = [(-np.inf, -1, -1), (-1, -1, 0), (-1, 0, 1),
                  (0, 1, 2), (1, 2, 5), (2, 5, 10)]
    starts = [(0, -0.91), (0, -1.12), (0, -1.13), (0, -.123),
              (0, 5.05), (0, 1.98)]
    ends = [(100, -0.95), (100, -1.05), (100, 1.2), (100, -0.123),
            (100, 5.04), (100, 1.96)]
    maxs = [(50, -5.02), (50, -0.053), (50, 0.122), (50, 1.051),
            (50, 1.973), (50, 5.48)]
    settings = prepare_test_settings(simtype='repptis')
    start = starts[example]
    end = ends[example]
    mxx = maxs[example]
    settings['simulation']['interfaces'] = interfaces[example]
    settings['simulation']['zero_ensemble'] = True
    settings['simulation']['flux'] = True
    settings['tis']['ensemble_number'] = example
    settings['tis']['detect'] = interfaces[example][-1]
    with patch('sys.stdout', new=StringIO()):
        ensemble = create_ensemble(settings)
    path = make_internal_path(start, end, mxx, ensemble['interfaces'][1])
    ensemble['path_ensemble'].add_path_data(path, 'ACC')
    return ensemble


def create_simple_system():
    """Create a simple system for testing."""
    box = create_box(periodic=[False])
    system = System(units='reduced', temperature=1.0, box=box)
    particles = Particles(dim=1)
    particles.add_particle(np.zeros((1, 1)), np.zeros((1, 1)),
                           np.zeros((1, 1)))
    system.particles = particles
    system.forcefield = ForceField('empty force field',
                                   potential=[MockPotential()])
    return system


def prepare_test_simulation_rnd():
    """Prepare a small system we can integrate."""
    settings = prepare_test_settings()
    settings['tis']['rgen'] = 'rgen'

    box = create_box(periodic=[False])
    system = System(temperature=settings['system']['temperature'],
                    units=settings['system']['units'], box=box)
    system.particles = Particles(dim=system.get_dim())

    system.forcefield = create_force_field(settings)
    system.add_particle(np.array([-1.0]), mass=1, name='Ar', ptype=0)
    check_engine(settings)
    engine = create_engine(settings)
    settings['system']['obj'] = system
    settings['engine']['obj'] = engine

    fill_up_tis_and_retis_settings(settings)
    check_ensemble(settings)
    with patch('sys.stdout', new=StringIO()):
        simulation = create_tis_simulation(settings)
    rgen = create_random_generator({'seed': 0})
    simulation.ensembles[0]['rgen'] = rgen
    simulation.ensembles[0]['system'] = system
    simulation.ensembles[0]['order_function'] = MockOrder()

    simulation.ensembles[0]['path_ensemble'].start_condition = 'L'
    system.particles.vel = np.array([[0.78008019924163818]])
    return simulation, settings


class TISTest(unittest.TestCase):
    """Run a test for the TIS algorithm.

    This painful test will compare with results obtained by the old
    FORTRAN TISMOL program.
    """

    def test_tis_001(self):
        """Test a TIS simulation for 001."""
        simulation, settings = prepare_test_simulation()
        ensemble = simulation.ensembles[0]
        # Magic tricks to have a backward compatibility
        ensemble['engine'].rgen = ensemble['rgen']
        ensemble['path_ensemble'].rgen = ensemble['engine'].rgen
        ensemble['path_ensemble'].start_condition = 'L'
        for i in range(10):
            with tempfile.TemporaryDirectory() as tempdir:
                make_dirs(os.path.join(tempdir, '002'))
                settings['ensemble'][0]['simulation']['exe_path'] = tempdir
                if i == 0:
                    with patch('sys.stdout', new=StringIO()):
                        for _ in initiate_path_simulation(simulation,
                                                          settings):
                            logging.debug('Running initialisation')
                else:
                    simulation.step()

            path = ensemble['path_ensemble'].paths[-1]
            path_data = [path['length'], path['status'], path['generated'][0],
                         path['ordermin'][0], path['ordermax'][0]]
            for j in (0, 1, 2):
                self.assertEqual(path_data[j], TISMOL_001[i][j])
            self.assertAlmostEqual(path_data[3], TISMOL_001[i][3], places=6)
            self.assertAlmostEqual(path_data[4], TISMOL_001[i][4], places=6)

    def test_tis_restart(self):
        """Test various TIS simulation restart."""
        sim1, settings1 = prepare_test_simulation_rnd()
        sim2, _ = prepare_test_simulation_rnd()
        sim3, settings3 = prepare_test_simulation_rnd()

        rgen1 = create_random_generator({'seed': 0})
        rgen2 = create_random_generator({'seed': 0})

        # This test is more a reminder than else. rgen1 != rgen2
        self.assertFalse(big_fat_comparer(rgen1, rgen2))

        with tempfile.TemporaryDirectory() as tempdir:
            make_dirs(os.path.join(tempdir, '002'))
            settings1['ensemble'][0]['simulation']['exe_path'] = tempdir
            settings3['ensemble'][0]['simulation']['exe_path'] = tempdir
            for i in range(11):
                if i == 0:
                    with patch('sys.stdout', new=StringIO()):
                        for _ in initiate_path_simulation(sim1, settings1):
                            logging.debug('Running initialisation')
                        for _ in initiate_path_simulation(sim3, settings3):
                            logging.debug('Running initialisation')
                else:
                    sim1.step()
                    sim3.step()

                lp1 = sim1.ensembles[0]['path_ensemble'].last_path
                lp3 = sim3.ensembles[0]['path_ensemble'].last_path

                pe1 = sim1.ensembles[0]['path_ensemble']
                pe3 = sim3.ensembles[0]['path_ensemble']
                self.assertTrue(lp1 == lp3)
                self.assertTrue(pe1 == pe3)

            sim3.step()

            lp1 = sim1.ensembles[0]['path_ensemble'].last_path
            lp3 = sim3.ensembles[0]['path_ensemble'].last_path

            with patch('sys.stdout', new=StringIO()):
                self.assertFalse(big_fat_comparer(lp1.restart_info(),
                                                  lp3.restart_info()))
            with self.assertRaises(ValueError):
                big_fat_comparer(lp1.restart_info(),
                                 lp3.restart_info(), hard=True)
            sim2.load_restart_info(copy.deepcopy(sim1.restart_info()))
            sim2.settings['ensemble'][0]['simulation']['exe_path'] = tempdir
            rst1 = {}
            rst1['order_function'] = sim1.ensembles[0]['order_function']
            rst1['path_ensemble'] = \
                sim1.ensembles[0]['path_ensemble'].restart_info()
            rst1['engine'] = sim1.ensembles[0]['engine'].restart_info()
            rst1['system'] = sim1.ensembles[0]['system'].restart_info()
            rst1['ens-rgen'] = sim1.ensembles[0]['rgen'].get_state()

            with patch('sys.stdout', new=StringIO()):
                sim2.ensembles[0]['engine'].load_restart_info(rst1['engine'])
                sim2.ensembles[0]['system'].load_restart_info(rst1['system'])
                sim2.ensembles[0]['rgen'].set_state(rst1['ens-rgen'])
                sim2.ensembles[0]['order_function'] = rst1['order_function']
                sim2.ensembles[0]['path_ensemble'].load_restart_info(
                    rst1['path_ensemble'])

            rst2 = sim2.ensembles[0]['rgen'].get_state()
            self.assertTrue(big_fat_comparer(rst1['ens-rgen'], rst2))

            lp1 = sim1.ensembles[0]['path_ensemble'].last_path
            lp2 = sim2.ensembles[0]['path_ensemble'].last_path
            self.assertEqual(lp1, lp2)
            # TODO: This is a hack: the force field is not loaded via
            # restart for systems. Previously when there were only one
            # system, it was assumed that the force field could just be
            # set up again from the input settings. Structurally, I don't
            # think the force field longer fits as a system attribute. It
            # is perhaps better as an ensemble attribute, given that all
            # phasepoints in an ensemble share the same force field?
            # Anyway, for sim2 to run below, the force field will have to
            # be set:
            # This is done now in the initiate restart for regular simulations
            for phasepoint in lp2.phasepoints:
                phasepoint.forcefield = sim2.ensembles[0]['system'].forcefield

            pe1 = sim1.ensembles[0]['path_ensemble']
            pe2 = sim2.ensembles[0]['path_ensemble']

            ps1, ps2 = pe1.paths[-1:][0], pe2.paths[-1:][0]
            ps2['cycle'] = ps1['cycle']
            pe2.paths = pe1.paths

            self.assertEqual(ps1, ps2)
            self.assertEqual(pe1, pe2)

            for i in range(5):
                sim2.step()

            lp2 = copy.deepcopy(sim2.ensembles[0]['path_ensemble'].last_path)
            pe2 = copy.deepcopy(sim2.ensembles[0]['path_ensemble'])

            sim3, settings3 = prepare_test_simulation_rnd()
            settings3['ensemble'][0]['simulation']['exe_path'] = tempdir
            for i in range(16):
                if i == 0:
                    with patch('sys.stdout', new=StringIO()):
                        for _ in initiate_path_simulation(sim3, settings3):
                            logging.debug('Running initialisation')
                else:
                    sim3.step()

        lp3 = sim3.ensembles[0]['path_ensemble'].last_path
        pe3 = sim3.ensembles[0]['path_ensemble']

        self.assertEqual(lp2, lp3)
        # Small hack to fix Rgen.nstate
        pe2.rgen = pe3.rgen
        self.assertEqual(pe2, pe3)


class TISTestMethod(unittest.TestCase):
    """Test the various TIS methods."""

    def test_time_reversal(self):
        """Test the time reversal move."""
        status = ('ACC', 'ACC', 'ACC', 'BWI')
        accept = (True, True, True, False)
        for i_case, (acc, stat) in enumerate(zip(accept, status)):
            ens = create_ensemble_and_paths(example=i_case)
            path = ens['path_ensemble'].last_path
            accept, new_path, status = time_reversal(
                path, SameOrder(), ens['path_ensemble'].interfaces,
                start_condition='L')
            for i, j in zip(path.phasepoints, reversed(new_path.phasepoints)):
                parti = i.particles
                partj = j.particles
                # Check that positions are the same:
                self.assertTrue(np.allclose(parti.pos, partj.pos))
                # Check that velocities are reversed:
                self.assertTrue(np.allclose(parti.vel, -1.0 * partj.vel))
                # Check that energies are the same:
                self.assertAlmostEqual(parti.ekin, partj.ekin)
                self.assertAlmostEqual(parti.vpot, partj.vpot)
                # Check that order parameters are the same:
                self.assertAlmostEqual(i.order[0], j.order[0])
            self.assertEqual(accept, acc)
            self.assertEqual(status, stat)
            self.assertEqual(new_path.get_move(), 'tr')
            # Check that the ld move is not altered:
            path.set_move('ld')
            accept, new_path, status = time_reversal(
                path, SameOrder(), ens['interfaces'], start_condition='L'
            )
            self.assertEqual(new_path.get_move(), 'ld')
            # Check that order parameters are reversed:
            accept, new_path, status = time_reversal(
                path, NegativeOrder(), ens['interfaces'], start_condition='L'
            )
            for i, j in zip(path.phasepoints,
                            reversed(new_path.phasepoints)):
                self.assertAlmostEqual(i.order[0], -1 * j.order[0])

    def test_shoot1(self):
        """Test the shooting move, vol 1."""
        ensemble = create_ensemble_and_paths(example=0)
        ensemble['order_function'] = MockOrder()
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()

        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False
        }
        # Generate BTX:
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)
        # Generate BTX with allowmaxlength:
        tis_settings['allowmaxlength'] = True
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)
        # Generate BWI:
        tis_settings['allowmaxlength'] = True
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = 1
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)
        # Generate FTL:
        tis_settings['allowmaxlength'] = False
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = -0.01
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'FTL')
        self.assertFalse(accept)

    def test_shoot2(self):
        """Test the shooting move, vol 2."""
        ensemble = create_ensemble_and_paths(example=2)
        ensemble['order_function'] = MockOrder()
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
        }

        # Generate NCR:
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = -0.1
        ensemble['interfaces'] = (-1., 9., 10.)
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'NCR')
        self.assertFalse(accept)

        # Generate FTL:
        ensemble['engine'] = MockEngine2(timestep=1.0,
                                         interfaces=ensemble['interfaces'])
        tis_settings['allowmaxlength'] = False
        ensemble['engine'].delta_v = -0.1
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'FTL')
        self.assertFalse(accept)

        # Generate FTX:
        ensemble['engine'] = MockEngine2(timestep=1.0,
                                         interfaces=ensemble['interfaces'])
        tis_settings['allowmaxlength'] = True
        ensemble['engine'].delta_v = -0.01
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'FTX')
        self.assertFalse(accept)

        # Generate FTX again - Test when move was 'ld':
        ensemble['engine'] = MockEngine2(timestep=1.0,
                                         interfaces=ensemble['interfaces'])
        tis_settings['allowmaxlength'] = False
        ensemble['path_ensemble'].last_path.set_move('ld')
        ensemble['engine'].delta_v = -0.01
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'FTX')
        self.assertFalse(accept)

    def test_shoot_repptis(self):
        """Test the shooting move for repptis."""
        ensemble = create_ensemble_and_paths_repptis(example=3)
        ensemble['order_function'] = MockOrder()
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()

        tis_settings = {
            'maxlength': 100,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False
        }

        start_cond = ensemble['path_ensemble'].start_condition
        # Generate BTX:
        accept, _, status = shoot(ensemble, tis_settings,
                                  start_cond=start_cond)
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)
        # Generate BTX with allowmaxlength:
        tis_settings['allowmaxlength'] = True
        accept, _, status = shoot(ensemble, tis_settings,
                                  start_cond=start_cond)
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)
        # Generate NCR:
        tis_settings['allowmaxlength'] = True
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = 1
        accept, _, status = shoot(ensemble, tis_settings,
                                  start_cond=start_cond)
        self.assertEqual(status, 'NCR')
        self.assertFalse(accept)
        # Generate FTL:
        tis_settings['allowmaxlength'] = False
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = -0.01
        accept, _, status = shoot(ensemble, tis_settings,
                                  start_cond=start_cond)
        self.assertEqual(status, 'BTL')
        self.assertFalse(accept)

    def test_shoot_kob(self):
        """Test the shooting move when we get a KOB."""
        ensemble = create_ensemble_and_paths()
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()
        ensemble['order_function'] = MockOrder2()
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False
        }

        # Generate BTL
        ensemble['rgen'] = MockRandomGenerator(seed=1)
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'KOB')
        self.assertFalse(accept)

    def test_shoot_aiming(self):
        """Test the non aimless shooting move."""
        ensemble = create_ensemble_and_paths(example=2)
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()
        ensemble['order_function'] = MockOrder()
        ensemble['rgen'] = MockRandomGenerator(seed=1)
        ensemble['path_ensemble'].rgen = MockRandomGenerator(seed=160)

        tis_settings = {'maxlength': 1000,
                        'sigma_v': -1,
                        'aimless': False,
                        'zero_momentum': False,
                        'rescale_energy': False,
                        'allowmaxlength': True}

        # Generate ACC
        ensemble['engine'].total_eclipse = float('inf')
        ensemble['engine'].delta_v = -0.01
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)
        # Generate MCR:
        accept, _, status = shoot(ensemble, tis_settings, start_cond='L')
        self.assertEqual(status, 'MCR')
        self.assertFalse(accept)

    def test_web_throwing_real(self):
        """Test web_throwing shooting move with real dynamics."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': True,
            'shooting_move': 'wt',
            'n_jumps': 22,
            'interface_sour': -1.95
        }

        interfaces = [-3, -1.4, 2.3]
        order_f = Position(index=0)
        engine = VelocityVerlet(0.01)

        trial_path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        shooting_point = trial_path.phasepoints[4]
        shooting_point.idx = 4
        shooting_point.dek = 0
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = trial_path
        ensemble = {'interfaces': interfaces, 'engine': engine,
                    'rgen': RandomGenerator(seed=0),
                    'path_ensemble': ens, 'order_function': order_f}

        _, path, _ = shoot(ensemble,
                           tis_settings,
                           start_cond='L',
                           shooting_point=shooting_point
                           )
        ens.last_path = path
        ensemble['interfaces'] = [-20, -1, 10]
        accept, _, status = web_throwing(ensemble, tis_settings)
        self.assertEqual(status, 'NSG')
        self.assertFalse(accept)

        ensemble['interfaces'] = [-3, -1.4, 2.3]
        accept, _, status = select_shoot(ensemble, tis_settings,
                                         start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        tis_settings['n_jumps'] = 0
        path.generated = ('ld', 0, 0, 0)
        ensemble['path_ensemble'].last_path = path.copy()
        accept, trial_path, status = web_throwing(ensemble, tis_settings)

        # With No jumps, the source path should be reproduced
        for ps1, ps2 in zip(trial_path.phasepoints, path.phasepoints):
            self.assertTrue(ps1 == ps2)  # This is an ESSENTIAL test
        for ps1, ps2 in zip(trial_path.phasepoints,
                            ensemble['path_ensemble'].last_path.phasepoints):
            self.assertTrue(ps1 == ps2)  # This is an ESSENTIAL test
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)
        self.assertEqual(trial_path.generated[0], 'ld')

        tis_settings['interface_sour'] = 10
        with self.assertRaises(AssertionError):
            web_throwing(ensemble, tis_settings)

        tis_settings['shooting_move'] = '00'
        accept, _, status = select_shoot(ensemble, tis_settings,
                                         start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

    def test_web_throwing(self):
        """Test web_throwing shooting move with rnd dynamics."""
        tis_settings = {
            'maxlength': 10000,
            'sigma_v': 10,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'n_jumps': 36,
            'shooting_move': 'wt',
            'interface_sour': 0.
        }
        interfaces = [-1., 2., 9.]
        order_f = Position(index=0)
        engine = RandomWalk(timestep=0.5)
        rgen = RandomGenerator(seed=16)

        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))

        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}
        accept, trial_path, status = shoot(ensemble, tis_settings,
                                           start_cond='L')

        # Check that the generated path is Ok
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        # Generate ACC:
        trial_path.generated = ('ld', 0, 0, 0)
        ensemble['path_ensemble'].last_path = trial_path
        accept, new_path, status = web_throwing(ensemble, tis_settings)
        self.assertEqual(new_path.generated[0], 'wt')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        # iF No jumps, the extension on this absurd interfaces
        # should get lost (BTX)
        tis_settings['n_jumps'] = 1
        ensemble['interfaces'] = [-3000000, 2, 300000]
        accept, new_path, status = select_shoot(ensemble, tis_settings,
                                                start_cond='L')
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)

        # should generate a B to B (thus a 'BWI')
        ensemble['interfaces'] = [-3000000, 2, 9]
        ensemble['engine'] = RandomWalk(timestep=1)
        accept, new_path, status = select_shoot(ensemble, tis_settings,
                                                start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

    def test_stone_skipping_fail(self):
        """Test stone skipping shooting move failures."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'ss',
            'n_jumps': 16,
        }
        # Generate ACC:
        order_f = Position(index=0)
        engine = MockEngine(16.0)
        rgen = RandomGenerator(seed=3)  # The seed helps the coverage.
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))

        ens = PathEnsemble(ensemble_number=1, interfaces=[0.7, 2., 2.5])
        path.set_move('ki')
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': [0.7, 2., 2.5]}
        accept, path, status = stone_skipping(ensemble, tis_settings,
                                              start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        order_f = Position(index=0)
        engine = MockEngine(16.0)
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))

        ens = PathEnsemble(ensemble_number=1, interfaces=[0.7, 2., 2.5])
        path.set_move('ss')
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': [0.7, 2., 2.5]}
        tis_settings['high_accept'] = False
        accept, path, status = stone_skipping(ensemble, tis_settings,
                                              start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        # B to B test -> Rejected
        ensemble['engine'] = VelocityVerlet(0.1)
        ensemble['interfaces'] = [0.7, 2., 1234.5]
        accept, path, status = stone_skipping(ensemble, tis_settings,
                                              start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        ensemble['engine'] = VelocityVerlet(0.001)
        ensemble['interfaces'] = [1110.7, 2., 1234.5]
        ensemble['rgen'].rand()
        # to let SS select the upper phasepoint in the only old_path crossing.
        accept, path, status = stone_skipping(ensemble, tis_settings,
                                              start_cond='L')
        self.assertEqual(status, 'XSS')
        self.assertFalse(accept)

        # Not extendable
        ensemble['path_ensemble'].last_path = path
        ensemble['interfaces'] = [-100., 1., 100.]
        accept, path, status = stone_skipping(ensemble, tis_settings,
                                              start_cond='L')
        self.assertEqual(status, 'NCR')
        self.assertFalse(accept)

    def test_stone_skipping_real(self):
        """Test stone skipping with real dynamics."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': True,
            'shooting_move': 'ss',
            'n_jumps': 26,
        }
        # Generate ACC:
        interfaces = [-3, -1.4, 2.3]
        order_f = Position(index=0)
        engine = VelocityVerlet(0.01)
        rgen = RandomGenerator(seed=5)

        trial_path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        shooting_point = trial_path.phasepoints[4]
        empty = trial_path.empty_path(rgen=RandomGenerator(seed=0),
                                      maxlen=1000)
        sub_ensemble = {'interfaces': interfaces,
                        'system': shooting_point.copy(),
                        'order_function': order_f}

        engine.propagate(empty, sub_ensemble, reverse=False)

        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        path = empty.reverse(order_f)
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}
        accept, path, status = select_shoot(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

    def test_stone_skipping_full_version(self):
        """Test stone skipping with real dynamics."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'ss',
            'n_jumps': 16,
            'high_accept': True,
        }
        # Generate ACC:
        interfaces = [-3, -1.4, 2.3]
        order_f = Position(index=0)
        engine = VelocityVerlet(0.01)
        rgen = RandomGenerator(seed=0)

        trial_path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        shooting_point = trial_path.phasepoints[4]
        empty = trial_path.empty_path(rgen=RandomGenerator(seed=0),
                                      maxlen=1000)
        sub_ensemble = {'interfaces': interfaces,
                        'system': shooting_point.copy(),
                        'order_function': order_f}

        engine.propagate(empty, sub_ensemble, reverse=False)

        path = empty.reverse(order_f)
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}
        accept, path, status = select_shoot(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        # Generate XSS - failed to do one stone skipping jump:
        tis_settings['maxlength'] = 32
        empty = trial_path.empty_path(rgen=RandomGenerator(seed=0),
                                      maxlen=1000)
        sub_ensemble = {'interfaces': interfaces,
                        'system': shooting_point.copy(),
                        'order_function': order_f}

        engine.propagate(empty, sub_ensemble, reverse=False)

        path = empty.reverse(order_f)
        ens.last_path = path
        accept, path, status = select_shoot(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'XSS')
        self.assertFalse(accept)

        # Generate NSS - failed to do one step crossing:
        ensemble['engine'] = VelocityVerlet(0.0000001)
        accept, path, status = select_shoot(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'NSS')
        self.assertFalse(accept)

    def test_wire_fencing_simple(self):
        """Test the common_sense move with MockEngine."""
        ensemble = create_ensemble_and_paths(example=0)
        ensemble['order_function'] = MockOrder()
        ensemble['engine'] = MockEngine(timestep=1.0)
        ensemble['system'] = create_simple_system()

        tis_settings = {
            'maxlength': 500,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'wf',
            'n_jumps': 6
        }
        ensemble['path_ensemble'].start_condition = 'L'
        ensemble['path_ensemble'].rgen = MockRandomGenerator(seed=0)
        ensemble['engine'].delta_v = 0.6
        accept, path, status = select_shoot(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        ensemble['path_ensemble'].last_path = path
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'WFA')
        self.assertFalse(accept)

        ensemble['engine'].delta_v = 0.012
        tis_settings['n_jumps'] = 2
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'NSG')
        self.assertFalse(accept)

        ensemble['interfaces'] = [-1.0, -0.9, -0.75]
        ensemble['engine'].total_eclipse = 30
        accept, _, status = wire_fencing(ensemble, tis_settings,
                                         start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        tis_settings['maxlength'] = 20
        accept, _, status = wire_fencing(ensemble, tis_settings,
                                         start_cond='L')
        self.assertEqual(status, 'FTX')
        self.assertFalse(accept)

    def test_wire_fencing_real(self):
        """Test common sense with real dynamics."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'wf',
            'n_jumps': 8,
            'high_accept': True,
        }
        # Generate ACC:
        interfaces = [-1.0, -0.5, 2.0]
        order_f = Position(index=0)
        engine = VelocityVerlet(0.01)
        rgen = RandomGenerator(seed=0)

        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path.set_move('wf')
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}

        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        tis_settings['high_accept'] = False
        ensemble['rgen'].rand()     # To fail detail balance
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'WFA')
        self.assertFalse(accept)

        ensemble['interfaces'] = [-1.0, 1.5, 2.0]
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        tis_settings['maxlength'] = 50
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)

        ensemble['interfaces'] = [-1.0, 2.5, 3.0]
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'NSG')
        self.assertFalse(accept)

    def test_wire_fencing_cap(self):
        """Test common sense with real dynamics."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'wf',
            'n_jumps': 6,
            'high_accept': True,
            'interface_cap': -0.4,
        }
        # Generate ACC:
        interfaces = [-1.0, -0.5, 2.0]
        order_f = Position(index=0)
        engine = VelocityVerlet(0.01)
        rgen = MockRandomGenerator(seed=0)

        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path.set_move('cs')
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = path
        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}

        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        tis_settings['high_accept'] = False
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'WFA')
        self.assertFalse(accept)

        tis_settings['maxlength'] = 50
        accept, path, status = wire_fencing(ensemble, tis_settings,
                                            start_cond='L')
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)

    def test_ss_rejected_detail_balance(self):
        """Test ss rejection with low and high acc."""
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'shooting_move': 'ss',
            'n_jumps': 5,
            'high_accept': False
        }
        # Generate ACC:
        trial_path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        interfaces = [-1, -0.4, 2.1]
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = trial_path
        ensemble = {'rgen': RandomGenerator(seed=13),
                    'engine': VelocityVerlet(0.1),
                    'order_function': Position(index=0),
                    'path_ensemble': ens,
                    'interfaces': interfaces}

        accept, _, status = select_shoot(ensemble, tis_settings,
                                         start_cond='L')

        self.assertEqual(status, 'BWI')
        self.assertFalse(accept)

        tis_settings['high_accept'] = True
        accept, _, status = select_shoot(ensemble, tis_settings,
                                         start_cond='L')

        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

    def test_ss_wt_wf_metropolis_acc_base(self):
        """Test ss_wt_wf_acceptance is working properly (part 0)."""
        # Generate a B to A path
        order_f = Position(index=0)
        interfaces = [-4, -2., -1.]
        lol_path = make_internal_path((5, 5), (-5, -5), (5, 5))
        new_path = lol_path.reverse()
        new_path.status = 'REJ'
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = lol_path
        ensemble = {'order_function': order_f,
                    'interfaces': interfaces,
                    'path_ensemble': ens}
        tis_settings = {'high_accept': True,
                        'shooting_move': 'ss'}
        succ = ss_wt_wf_metropolis_acc(path_new=new_path,
                                       ensemble=ensemble,
                                       tis_settings=tis_settings)
        self.assertTrue(succ)
        self.assertEqual(new_path.status, 'ACC')
        succ, trial_path = ss_wt_wf_acceptance(trial_path=new_path,
                                               ensemble=ensemble,
                                               tis_settings=tis_settings,
                                               start_cond='R')
        self.assertTrue(succ)
        self.assertEqual(new_path.phasepoints[0].order,
                         trial_path.phasepoints[-1].order)

    def test_ss_wt_wf_metropolis_acc(self):
        """Test ss_wt_wf_acceptance is working properly."""
        order_f = Position(index=0)
        engine = RandomWalk(timestep=0.1)
        interfaces = [-1, 0., 1]
        rgen = RandomGenerator(seed=0)
        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)

        ensemble = {'rgen': rgen, 'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}
        # Generate a B to A path
        path = make_internal_path((5, 5), (-5, -5), (5, 5))

        # Fix it and accept it
        interfaces = [-4, -2., -1]
        path2 = make_internal_path((-5, -5), (-5, -5), (5, 5))
        old_path = paste_paths(path2, path)
        ensemble['interfaces'] = interfaces
        ensemble['path_ensemble'].last_path = old_path
        succ = ss_wt_wf_metropolis_acc(path_new=path,
                                       ensemble=ensemble,
                                       tis_settings={'high_accept': True,
                                                     'shooting_move': 'ss'})
        self.assertFalse(succ)
        self.assertEqual(path.status, 'BWI')

        ensemble['rgen'] = RandomGenerator(seed=0)
        ensemble['path_ensemble'].last_path = path
        ensemble['interfaces'] = interfaces
        succ = ss_wt_wf_metropolis_acc(path_new=old_path,
                                       ensemble=ensemble,
                                       tis_settings={'shooting_move': 'ss'})
        self.assertFalse(succ)
        self.assertEqual(old_path.status, 'SSA')

        ensemble['path_ensemble'].last_path = path
        succ = ss_wt_wf_metropolis_acc(path_new=old_path,
                                       ensemble=ensemble,
                                       tis_settings={'interface_sour': -2.,
                                                     'shooting_move': 'wt'})
        self.assertFalse(succ)
        self.assertEqual(old_path.status, 'WTA')

        ensemble['path_ensemble'].last_path = path
        succ = ss_wt_wf_metropolis_acc(path_new=old_path,
                                       ensemble=ensemble,
                                       tis_settings={'high_accept': True,
                                                     'shooting_move': 'ss'})
        self.assertTrue(succ)
        self.assertEqual(old_path.status, 'ACC')

    def test_extender(self):
        """Test extender of paths."""
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path.status = 'ACC'
        order_f = Position(index=0)
        engine = MockEngine(2.0)
        interfaces = [-1, 0., 2.]
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': False,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
        }

        ens = PathEnsemble(ensemble_number=1, interfaces=interfaces)
        ens.last_path = path
        ensemble = {'engine': engine, 'order_function': order_f,
                    'path_ensemble': ens, 'interfaces': interfaces}
        # Check that nothing changes, the path is already good
        accept, new_path, status = extender(path, ensemble, tis_settings)
        self.assertTrue(new_path == path)
        self.assertTrue(accept)

        # Check that the forward expansion works, and it can fail
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path.status = 'ACC'

        interfaces = [-1, 0., 2.7]
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': False,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
        }
        ensemble['path_ensemble'].last_path = path
        accept, new_path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        ensemble['interfaces'] = [-1, 0., 3.]
        accept, new_path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'FTX')
        self.assertFalse(accept)

        ensemble['interfaces'] = [2., 2.5, 3.]
        accept, new_path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

    def test_extender2(self):
        """Test extender behaviour in special circumstances."""
        path = make_internal_path((-1, 2.1), (-2, 2.2), (-1, -1))
        path.status = 'ACC'
        order_f = Position(index=0)
        engine = MockEngine(2.0)
        # Test extension in the 000^- ensemble
        interfaces = [-9999., 0., 0.]
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': False,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
            'n_jumps': 16,
            'interface_sour': 1.5
        }
        # Check that nothing changes, the path is already good

        ensemble = {'engine': engine, 'order_function': order_f,
                    'interfaces': interfaces}
        accept, _, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        # Test segment direction
        # (just wrong orientation but still connecting B to A)
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path_new = path.reverse(order_f)
        path_new.maxlen = 1000
        ensemble['interfaces'] = [-1, 0., 2.4]
        accept, path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        accept, path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'ACC')
        self.assertTrue(accept)

        tis_settings['maxlength'] = 10
        accept, path, status = extender(path, ensemble, tis_settings)
        self.assertEqual(status, 'FTX')
        self.assertFalse(accept)

        # Test segment direction, shall fail
        path = make_internal_path((-5, 2.1), (5, 2.2), (1, -1))
        path_new = path.reverse(order_f)
        path_new.maxlen = 1000

        accept, path, status = extender(path_new, ensemble, tis_settings)
        self.assertEqual(status, 'BTX')
        self.assertFalse(accept)

    def test_0_min_left(self):
        """Test that a path that ends up at the left interface of the
        {0-} ensemble is rejected."""
        interfaces = [0, 1, 2]
        ens = PathEnsemble(ensemble_number=0, interfaces=interfaces)
        path = make_internal_path((0, 2.1), (100, 2.2), (50, -1),
                                  ens.interfaces[1])
        ens.add_path_data(path, status='ACC')
        order_f = MockOrder()
        engine = MockEngine(10)
        ens.last_path = path
        ensemble = {'engine': engine,
                    'order_function': order_f,
                    'rgen': path.rgen,
                    'path_ensemble': ens,
                    'interfaces': interfaces}
        with tempfile.TemporaryDirectory() as tempdir:
            settings = {'tis': {'freq': 0, 'maxlength': 1000, 'sigma_v': -1,
                                'aimless': True, 'zero_momentum': False,
                                'rescale_energy': False,
                                'allowmaxlength': False},
                        'simulation': {'task': 'tis',
                                       'exe_path': tempdir}}
            make_dirs(os.path.join(tempdir, '000'))
            out = make_tis_step_ensemble(ensemble, settings, cycle=1)

        # Make sure this generated an L to R path
        self.assertGreater(out[1].phasepoints[0].order[0], 2)
        self.assertLess(out[1].phasepoints[-1].order[0], 0)

        # Assert if this was rejected with the right status
        self.assertFalse(out[0])
        self.assertEqual(out[2], '0-L')

    def test_0_min_right(self):
        """Test that a path that ends up at the right interface of the
        {0-} ensemble is accepted"""
        interfaces = [0, 1, 2]
        ens = PathEnsemble(ensemble_number=0, interfaces=interfaces)
        path = make_internal_path((0, 2.1), (100, 2.2), (50, -1),
                                  ens.interfaces[1])
        ens.add_path_data(path, status='ACC')
        order_f = MockOrder()
        engine = MockEngine(10, turn_around=2000)

        ens.last_path = path
        ensemble = {'engine': engine,
                    'order_function': order_f,
                    'rgen': path.rgen,
                    'path_ensemble': ens,
                    'interfaces': interfaces}

        with tempfile.TemporaryDirectory() as tempdir:
            settings = {'tis': {'freq': 0,
                                'maxlength': 1000,
                                'sigma_v': -1,
                                'aimless': True,
                                'zero_momentum': False,
                                'rescale_energy': False,
                                'allowmaxlength': False},
                        'simulation': {'task': 'tis',
                                       'exe_path': tempdir}}

            make_dirs(os.path.join(tempdir, '000'))
            out = make_tis_step_ensemble(ensemble, settings, cycle=1)

        # Make sure this generated an R to R path
        self.assertGreater(out[1].phasepoints[0].order[0], 2)
        self.assertGreater(out[1].phasepoints[-1].order[0], 2)

        # Assert if this was accepted with the right status
        self.assertTrue(out[0])
        self.assertEqual(out[2], 'ACC')

    def test_make_tis(self):
        """Test that we can do the RETIS steps."""
        settings, ensembles = create_ensembles_and_paths()
        ens_l = ['000', '001', '002', '003', '004']
        rgen = MockRandomGenerator(seed=0)
        for setting in settings['ensemble']:
            setting['tis']['freq'] = 1
        # Check that we can do RETIS:
        settings['tis']['nullmoves'] = True
        ensembles[1]['path_ensemble'].last_path.set_move('ld')
        with tempfile.TemporaryDirectory() as tempdir:
            for i, fld in enumerate(ens_l):
                make_dirs(os.path.join(tempdir, fld))
                settings['ensemble'][i]['simulation']['exe_path'] = tempdir
            results = make_tis(ensembles, rgen, settings, cycle=0)
            for _ in results:
                continue
        self.assertTrue(ensembles[1]['path_ensemble'].last_path.get_move() ==
                        'ld')

        # Check that we can do relative shoots:
        settings['tis']['relative_shoots'] = [0.1, 0.1, 0.1, 0.1, 0.6]
        with tempfile.TemporaryDirectory() as tempdir:
            for i, fld in enumerate(ens_l):
                make_dirs(os.path.join(tempdir, fld))
                settings['ensemble'][i]['simulation']['exe_path'] = tempdir
            settings['simulation']['exe_path'] = tempdir
            results = make_tis(ensembles, rgen, settings, cycle=1)
            moves = ('tr', 'ld', 'tr', 'tr', 'tr')
            for resi in results:
                self.assertEqual(
                    resi['mc-move'], moves[resi['ensemble_number']]
                )

    def test_explorer(self):
        """Test that we can use the explore approach."""
        settings, ensembles = create_ensembles_and_paths()
        ens_l = ['000', '001', '002', '003', '004']
        rgen = MockRandomGenerator(seed=0)

        # Check that we can explore:
        with tempfile.TemporaryDirectory() as tempdir:
            for i, fld in enumerate(ens_l):
                make_dirs(os.path.join(tempdir, fld))
                settings['ensemble'][i]['simulation']['exe_path'] = tempdir
                settings['ensemble'][i]['simulation']['task'] = 'explore'
                settings['ensemble'][i]['tis']['freq'] = 0

            results = make_tis(ensembles, rgen, settings, cycle=0)
            for _ in results:
                continue

        self.assertEqual(ensembles[1]['path_ensemble'].last_path.status, 'EXP')


class TestPermMethods(unittest.TestCase):
    def setUp(self):
        self.settings, ensembles = create_ensembles_and_paths()
        self.settings['tis']['maxlength'] = 1000
        self.ens = ensembles[0]
        self.ens['interfaces'][0] = -2.0

        self.ens['path_ensemble'].start_cond = set(['L', 'R'])

    def test_to_many_weights(self):
        # Set settings to something crazy
        self.settings['tis']['freq'] = 2
        with self.assertRaises(ValueError):
            make_tis_step_ensemble(self.ens, self.settings, 0)

    def test_breaking_rgen(self):
        class BrokenChoiceRgen:
            def choice(self, _, __, p):
                """Broken function."""
                return f'Broken {p}'
        broken_choice = BrokenChoiceRgen()
        self.ens['rgen'] = broken_choice
        with self.assertRaises(RuntimeError):
            make_tis_step_ensemble(self.ens, self.settings, 0)

    def test_mirror_selection(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['mirror_freq'] = 1
        path = self.ens['path_ensemble'].last_path
        accept, trial, _ = make_tis_step_ensemble(self.ens,
                                                  self.settings, 1)
        assert accept is True
        assert trial.status == 'ACC'
        for i, j in zip(path.phasepoints, trial.phasepoints):
            # Mirrored mock order is just mirrored around 0
            assert i.order[0] == -j.order[0]
        assert trial.generated[0] == 'mr'

    def test_mirror_dont_override_ld(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['mirror_freq'] = 1
        self.ens['path_ensemble'].last_path.set_move('ld')

        accept, trial, _ = make_tis_step_ensemble(self.ens, self.settings, 1)
        assert accept is True
        assert trial.status == 'ACC'
        assert trial.generated[0] == 'ld'

    def test_no_valid_target_swap(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        # Target index is the same as op.index, so no valid points to swap to
        self.settings['tis']['target_indices'] = [1]
        self.ens['order_function'].index = 1
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is False
        assert status == "TSS"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 1

    def test_current_index_not_in_target_indices(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        # Target index is the same as op.index, so no valid points to swap to
        self.settings['tis']['target_indices'] = [1]
        self.ens['order_function'].index = 0
        with self.assertRaisesRegex(ValueError, "0, is not part of the target "
                                    r"indices: \[1\]"
                                    ):
            make_tis_step_ensemble(self.ens, self.settings, 1)

    def test_mc_rejection(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [0, 1]
        self.ens['order_function'].index = 1
        # Here we hack the random rgen to give a number above 1 at the right
        # time (documented incase this order changes)
        # 0 = select move
        # 1 = select index to shoot
        # 2 = determine acceptance <= alter this one
        self.ens['rgen'].rgen[2] = 1.1
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is False
        assert status == "TSA"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 1

    def test_btx_rejection(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [0, 1]
        self.ens['order_function'].index = 1
        # Due to the mock rgen we know the backward path will be length 2
        self.settings['tis']['maxlength'] = 2
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is False
        assert status == "BTX"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 1

    def test_ftx_rejection(self):
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [0, 1]
        self.ens['order_function'].index = 1
        # Due to the mock rgen we know the backward path will be length 2
        # Forward will be 9, so test 3 to see if we run, but still end with FTX
        self.settings['tis']['maxlength'] = 3
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is False
        assert status == "FTX"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 1

    def test_not_all_bw_frames_are_valid(self):
        # Normal path will trigger forward cut, try backwards instead
        path = make_internal_path((0, -2.1), (100, -1.0), (100, -1.0))

        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [0, 1]
        self.ens['order_function'].index = 1
        self.ens['path_ensemble'].last_path = path
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is True
        assert status == "ACC"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 0

    def test_identity_target_swap(self):
        # index is not used in op, so this should return an identical path
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [1, 0]
        self.ens['order_function'].index = 1
        path = self.ens['path_ensemble'].last_path
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        assert accept is True
        assert status == "ACC"
        assert trial.generated[0] == 'ts'
        assert self.ens['order_function'].index == 0
        for i, j in zip(path.phasepoints, trial.phasepoints):
            assert i.order == j.order

    def test_m_pass_for_l_r_start_cond(self):
        path = make_internal_path((0, -2.01), (100, -2.01), (50, -1.01))
        self.ens['path_ensemble'].start_condition = ("R", "L")
        self.ens['path_ensemble'].last_path = path
        self.ens['engine'] = MockEngine(-10)

        self.settings['tis'] = {'freq': 0,
                                'maxlength': 1000,
                                'sigma_v': -1,
                                'aimless': True,
                                'zero_momentum': False,
                                'rescale_energy': False,
                                'allowmaxlength': False}

        _, _, status = shoot(self.ens, self.settings['tis'], ('L', 'R'))
        assert status == 'ACC'

    def test_bw_extend(self):
        # Start with an incomplete path so we must extend
        path = make_internal_path((0, -1.9), (100, -1.1), (100, -1.1))
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [1, 0]
        self.ens['order_function'].index = 1
        self.ens['path_ensemble'].last_path = path
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        # Due to mock engine this will be BTX
        assert accept is False
        assert status == "BTX"
        assert trial.length > path.length
        assert self.ens['order_function'].index == 1

    def test_fw_extend(self):
        # Start with an incomplete path so we must extend
        path = make_internal_path((0, -2.001), (100, -1.1), (100, -1.1))
        self.settings['tis']['freq'] = 0
        self.settings['tis']['target_freq'] = 1
        self.settings['tis']['target_indices'] = [1, 0]
        self.ens['order_function'].index = 1
        self.ens['path_ensemble'].last_path = path
        accept, trial, status = make_tis_step_ensemble(self.ens,
                                                       self.settings, 1)
        # Due to mock engine this will be FTX
        assert accept is True
        assert status == "ACC"
        # Make sure we actually did md
        assert trial.length > path.length
        assert self.ens['order_function'].index == 0


if __name__ == '__main__':
    unittest.main()
